"""
This file is the entry point for the python get model command for the package

TODO: use selector and select for output

"""
import atexit
import json
import logging
import os
import sys
from contextlib import ExitStack

import numba as nb
import numpy as np
import pandas as pd

from numba import int32 as nb_int32
from numba.typed import Dict
from numba.types import Tuple as nb_Tuple

from oasis_data_manager.df_reader.config import get_df_reader, clean_config, InputReaderConfig
from oasis_data_manager.filestore.backends.base import BaseStorage
from oasis_data_manager.filestore.config import get_storage_from_config_path
from oasislmf.pytools.common.event_stream import PIPE_CAPACITY
from oasislmf.utils.data import validate_vulnerability_replacements, analysis_settings_loader
from oasislmf.pytools.common.data import (
    areaperil_int, areaperil_int_size,
    oasis_float, oasis_float_size,
    oasis_int, oasis_int_size,
    damagebin_dtype, items_dtype,
    vulnerability_dtype
)
from oasislmf.pytools.data_layer.footprint_layer import FootprintLayerClient
from oasislmf.pytools.getmodel.common import Index_type, Keys
from oasislmf.pytools.getmodel.footprint import Footprint
from oasislmf.pytools.utils import redirect_logging
from oasislmf.utils.ping import oasis_ping

from .vulnerability import vulnerability_dataset, parquetvulnerability_meta_filename
from ..common.data import null_index

logger = logging.getLogger(__name__)

buff_size = PIPE_CAPACITY

buff_int_size = buff_size // oasis_int_size

areaperil_int_relative_size = areaperil_int_size // oasis_int_size
oasis_float_relative_size = oasis_float_size // oasis_int_size
results_relative_size = 2 * oasis_float_relative_size

VulnerabilityIndex_dtype = np.dtype([
    ('vulnerability_id', np.int32),
    ('offset', np.int64),
    ('size', np.int64),
    ('original_size', np.int64)
])
VulnerabilityIndex = nb.from_dtype(VulnerabilityIndex_dtype)

VulnerabilityRow_dtype = np.dtype([
    ('intensity_bin_id', np.int32),
    ('damage_bin_id', np.int32),
    ('probability', oasis_float)
])
VulnerabilityRow = nb.from_dtype(VulnerabilityRow_dtype)

vuln_offset = 4


@nb.njit(cache=True)
def load_areaperil_id_u4(int32_mv, cursor, areaperil_id):
    int32_mv[cursor] = areaperil_id.view('i4')
    return cursor + 1


@nb.njit(cache=True)
def load_areaperil_id_u8(int32_mv, cursor, areaperil_id):
    int32_mv[cursor: cursor + 1] = areaperil_id.view('i4')
    return cursor + 2


if areaperil_int == 'u4':
    load_areaperil_id = load_areaperil_id_u4
elif areaperil_int == 'u8':
    load_areaperil_id = load_areaperil_id_u8
else:
    raise Exception(f"AREAPERIL_TYPE {areaperil_int} is not implemented chose u4 or u8")


@nb.njit(cache=True)
def load_items(items, valid_area_peril_id):
    """
    Processes the Items loaded from the file extracting meta data around the vulnerability data.

    Args:
        items: (List[item_dtype]) Data loaded from the vulnerability file
        valid_area_peril_id: array of area_peril_id to be included (if none, all are included)

    Returns: (Tuple[Dict[int, int], List[int], Dict[int, int], List[Tuple[int, int]], List[int]])
             vulnerability dictionary, vulnerability IDs, areaperil to vulnerability index dictionary,
             areaperil ID to vulnerability index array, areaperil ID to vulnerability array
    """
    areaperil_to_vulns_size = 0
    areaperil_dict = Dict()
    vuln_dict = Dict()
    vuln_idx = 0
    for i in range(items.shape[0]):
        item = items[i]

        # filter areaperil_id
        if valid_area_peril_id is not None and item['areaperil_id'] not in valid_area_peril_id:
            continue

        # insert the vulnerability index if not in there
        if item['vulnerability_id'] not in vuln_dict:
            vuln_dict[item['vulnerability_id']] = np.int32(vuln_idx)
            vuln_idx += 1

        # insert an area dictionary into areaperil_dict under the key of areaperil ID
        if item['areaperil_id'] not in areaperil_dict:
            area_vuln = Dict()
            area_vuln[item['vulnerability_id']] = 0
            areaperil_dict[item['areaperil_id']] = area_vuln
            areaperil_to_vulns_size += 1
        else:
            if item['vulnerability_id'] not in areaperil_dict[item['areaperil_id']]:
                areaperil_to_vulns_size += 1
                areaperil_dict[item['areaperil_id']][item['vulnerability_id']] = 0

    areaperil_to_vulns_idx_dict = Dict()
    areaperil_to_vulns_idx_array = np.empty(len(areaperil_dict), dtype=Index_type)
    areaperil_to_vulns = np.empty(areaperil_to_vulns_size, dtype=np.int32)

    areaperil_i = 0
    vulnerability_i = 0

    for areaperil_id, vulns in areaperil_dict.items():
        areaperil_to_vulns_idx_dict[areaperil_id] = areaperil_i
        areaperil_to_vulns_idx_array[areaperil_i]['start'] = vulnerability_i

        for vuln_id in sorted(vulns):  # sorted is not necessary but doesn't impede the perf and align with cpp getmodel
            areaperil_to_vulns[vulnerability_i] = vuln_id
            vulnerability_i += 1
        areaperil_to_vulns_idx_array[areaperil_i]['end'] = vulnerability_i
        areaperil_i += 1

    return vuln_dict, areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns


def get_items(input_path, ignore_file_type=set(), valid_area_peril_id=None):
    """
    Loads the items from the items file.

    Args:
        input_path: (str) the path pointing to the file
        ignore_file_type: set(str) file extension to ignore when loading

    Returns: (Tuple[Dict[int, int], List[int], Dict[int, int], List[Tuple[int, int]], List[int]])
             vulnerability dictionary, vulnerability IDs, areaperil to vulnerability index dictionary,
             areaperil ID to vulnerability index array, areaperil ID to vulnerability array
    """
    input_files = set(os.listdir(input_path))
    if "items.bin" in input_files and "bin" not in ignore_file_type:
        logger.debug(f"loading {os.path.join(input_path, 'items.csv')}")
        items = np.memmap(os.path.join(input_path, "items.bin"), dtype=items_dtype, mode='r')
    elif "items.csv" in input_files and "csv" not in ignore_file_type:
        logger.debug(f"loading {os.path.join(input_path, 'items.csv')}")
        items = np.loadtxt(os.path.join(input_path, "items.csv"), dtype=items_dtype, delimiter=",", skiprows=1, ndmin=1)
    else:
        raise FileNotFoundError(f'items file not found at {input_path}')

    return load_items(items, valid_area_peril_id)


def get_intensity_bin_dict(input_path):
    """
    Loads the intensity bin dictionary file and creates a dictionary to map intensities to bins
    Used in the dynamic footprint generation as intensitys can be adjusted for defences at runtime

    Args:
        input_path: (str) the path pointing to the file

    Returns: (Dict[(int, int), int])
             intensity bin dict,
             with index of peril_id (encoded) and intensity value, and value of bin index
    """
    input_files = set(os.listdir(input_path))
    intensity_bin_dict = Dict.empty(nb_Tuple((nb_int32, nb_int32)), nb_int32)
    if "intensity_bin_dict.csv" in input_files:
        logger.debug(f"loading {os.path.join(input_path, 'intensity_bin_dict.csv')}")
        data = pd.read_csv(os.path.join(input_path, "intensity_bin_dict.csv"))
        data = data[['peril_id', 'intensity', 'intensity_bin']]
        data['peril_id'] = data['peril_id'].apply(encode_peril_id)
        data = data.to_records(index=False).tolist()
        data = np.array(data, dtype=np.int32)
        for d in data:
            intensity_bin_dict[(d[0], d[1])] = d[2]
    else:
        raise FileNotFoundError(f'intensity_bin_dict file not found at {input_path}')

    return intensity_bin_dict


def encode_peril_id(peril_id):
    """Encode a string to an integer.

    Args:
        peril_id (str): 3-digit Oasis peril code (also works with numeric codes).

    Returns:
        int: The encoded peril_id.
    """

    return sum(ord(c) << (8 * i) for i, c in enumerate(str(peril_id).upper()))


def get_intensity_adjustment(input_path):
    pass


@nb.njit(cache=True)
def load_vuln_probability(vuln_array, vuln, vuln_id):
    if vuln_array.shape[0] < vuln['damage_bin_id']:
        raise Exception("vulnerability_id " + str(vuln_id) + " has damage_bin_id bigger that expected maximum")
    if vuln['intensity_bin_id'] <= vuln_array.shape[1]:  # intensity in vulnerability curve but not in the footprint, we can ignore
        vuln_array[vuln['damage_bin_id'] - 1, vuln['intensity_bin_id'] - 1] = vuln['probability']


@nb.njit(cache=True)
def load_vulns_bin_idx(vulns_bin, vulns_idx_bin, vuln_dict,
                       num_damage_bins, num_intensity_bins, rowsize):
    """
    Loads the vulnerability binary index file.

    Args:
        vulns_bin: (List[VulnerabilityRow]) vulnerability data from the vulnerability file
        vulns_idx_bin: (List[VulnerabilityIndex]) vulnerability index data from the vulnerability idx file
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_damage_bins: (int) number of damage bins in the data
        num_intensity_bins: (int) the number of intensity bins

    Returns: (List[List[List[floats]]]) vulnerability data grouped by intensity bin and damage bin
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins), dtype=oasis_float)
    vuln_ids = np.full(len(vuln_dict), null_index)
    for idx_i in range(vulns_idx_bin.shape[0]):
        vuln_idx = vulns_idx_bin[idx_i]
        if vuln_idx['vulnerability_id'] in vuln_dict:
            vuln_ids[vuln_dict[vuln_idx['vulnerability_id']]] = vuln_idx['vulnerability_id']
            cur_vuln_array = vuln_array[vuln_dict[vuln_idx['vulnerability_id']]]
            start = (vuln_idx['offset'] - vuln_offset) // rowsize
            end = start + vuln_idx['size'] // rowsize
            for vuln_i in range(start, end):
                vuln = vulns_bin[vuln_i]
                load_vuln_probability(cur_vuln_array, vuln, vuln_idx['vulnerability_id'])

    return vuln_array, vuln_ids


@nb.njit(cache=True)
def load_vulns_bin_idx_adjusted(vulns_bin, vulns_idx_bin, vuln_dict,
                                num_damage_bins, num_intensity_bins, rowsize, adj_vuln_data=None):
    """
    Loads the vulnerability binary index file, prioritizing the data in the adjustments file over the data in the
    vulnerability file.

    Args:
        vulns_bin: (List[VulnerabilityRow]) vulnerability data from the vulnerability file
        vulns_idx_bin: (List[VulnerabilityIndex]) vulnerability index data from the vulnerability idx file
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_damage_bins: (int) number of damage bins in the data
        num_intensity_bins: (int) the number of intensity bins
        adj_vuln_data: (List[vulnerability_dtype]) vulnerability adjustment data, sorted by vuln_id

    Returns: (List[List[List[floats]]]) vulnerability data grouped by intensity bin and damage bin
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins), dtype=oasis_float)
    vuln_ids = np.full(len(vuln_dict), null_index)
    adj_vuln_index = 0

    for idx_i in range(vulns_idx_bin.shape[0]):
        vuln_idx = vulns_idx_bin[idx_i]
        vuln_id = vuln_idx['vulnerability_id']

        # Check if current vulnerability id is in the adjustment data
        while adj_vuln_data is not None and adj_vuln_index < len(adj_vuln_data) and adj_vuln_data[adj_vuln_index]['vulnerability_id'] < vuln_id:
            adj_vuln_index += 1

        if vuln_id in vuln_dict:
            vuln_ids[vuln_dict[vuln_id]] = vuln_id
            cur_vuln_array = vuln_array[vuln_dict[vuln_id]]
            start = (vuln_idx['offset'] - vuln_offset) // rowsize
            end = start + vuln_idx['size'] // rowsize

            # Apply data from vulns_bin or adj_vuln_data
            for vuln_i in range(start, end):
                if (adj_vuln_data is not None and adj_vuln_index < len(adj_vuln_data) and adj_vuln_data[adj_vuln_index]['vulnerability_id'] == vuln_id):
                    load_vuln_probability(cur_vuln_array, adj_vuln_data[adj_vuln_index], vuln_id)
                    adj_vuln_index += 1
                else:
                    load_vuln_probability(cur_vuln_array, vulns_bin[vuln_i], vuln_id)

    # Add remaining adj_vuln_data
    while adj_vuln_data is not None and adj_vuln_index < len(adj_vuln_data):
        adj_vuln = adj_vuln_data[adj_vuln_index]
        vuln_id = adj_vuln['vulnerability_id']
        if vuln_id in vuln_dict:
            load_vuln_probability(vuln_array[vuln_dict[vuln_id]], adj_vuln, vuln_id)
        adj_vuln_index += 1

    return vuln_array, vuln_ids


@nb.njit(cache=True)
def load_vulns_bin(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins):
    """
    Loads the vulnerability data grouped by the intensity and damage bins.

    Args:
        vuln_bin: (List[Vulnerability]) vulnerability data from the vulnerability file
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_damage_bins: (int) number of damage bins in the data
        num_intensity_bins: (int) the number of intensity bins

    Returns: (List[List[List[floats]]]) vulnerability data grouped by intensity bin and damage bin
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins), dtype=oasis_float)
    vuln_ids = np.full(len(vuln_dict), null_index)
    cur_vulnerability_id = -1

    for vuln_i in range(vulns_bin.shape[0]):
        vuln = vulns_bin[vuln_i]
        if vuln['vulnerability_id'] != cur_vulnerability_id:
            if vuln['vulnerability_id'] in vuln_dict:
                cur_vulnerability_id = vuln['vulnerability_id']
                vuln_ids[vuln_dict[cur_vulnerability_id]] = cur_vulnerability_id
                cur_vuln_array = vuln_array[vuln_dict[cur_vulnerability_id]]
            else:
                cur_vulnerability_id = -1
        if cur_vulnerability_id != -1:
            load_vuln_probability(cur_vuln_array, vuln, cur_vulnerability_id)

    return vuln_array, vuln_ids


@nb.njit(cache=True)
def load_vulns_bin_adjusted(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins, adj_vuln_data=None):
    """
    Loads the vulnerability data grouped by the intensity and damage bins, prioritizing the data
    in the adjustments file over the data in the vulnerability file.

    Args:
        vuln_bin: (List[Vulnerability]) vulnerability data from the vulnerability file
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_damage_bins: (int) number of damage bins in the data
        num_intensity_bins: (int) the number of intensity bins
        adj_vuln_data: (List[vulnerability_dtype]) vulnerability adjustment data, sorted by vuln_id

    Returns: (List[List[List[floats]]]) vulnerability data grouped by intensity bin and damage bin
    """
    vuln_array = np.zeros((len(vuln_dict), num_damage_bins, num_intensity_bins), dtype=oasis_float)
    vuln_ids = np.full(len(vuln_dict), null_index)
    ids_to_replace = set()
    adj_vuln_index = 0

    # Create list of ids to replace if adj_vuln_data is provided
    if adj_vuln_data is not None:
        for adj_vuln in adj_vuln_data:
            ids_to_replace.add(adj_vuln['vulnerability_id'])

    vuln_i = 0

    while vuln_i < len(vulns_bin):
        vuln = vulns_bin[vuln_i]
        vuln_id = vuln['vulnerability_id']
        if vuln_id in vuln_dict:
            vuln_ids[vuln_dict[vuln_id]] = vuln_id
            if vuln_id in ids_to_replace:
                # Advance to current vuln_id
                while adj_vuln_index < len(adj_vuln_data) and adj_vuln_data[adj_vuln_index]['vulnerability_id'] < vuln_id:
                    adj_vuln_index += 1
                # process current vuln_id
                while adj_vuln_index < len(adj_vuln_data) and adj_vuln_data[adj_vuln_index]['vulnerability_id'] == vuln_id:
                    load_vuln_probability(vuln_array[vuln_dict[vuln_id]], adj_vuln_data[adj_vuln_index], vuln_id)
                    adj_vuln_index += 1
                # Skip remaining vulns_bin entries with the same vulnerability_id
                while vuln_i < len(vulns_bin) and vulns_bin[vuln_i]['vulnerability_id'] == vuln_id:
                    vuln_i += 1
                continue
            else:
                # Use data from vulns_bin
                load_vuln_probability(vuln_array[vuln_dict[vuln_id]], vuln, vuln_id)
        vuln_i += 1
    return vuln_array, vuln_ids


@nb.njit(cache=True)
def update_vulns_dictionary(vuln_dict, vulns_id_array):
    """
    Updates the indexes of the vulnerability IDs (usually used in loading vulnerability data from parquet file).

    Args:
        vuln_dict: (Dict[int, int]) vulnerability dict that maps the vulnerability IDs (key) with the index (value)
        vulns_id_array: (List[int]) list of vulnerability IDs loaded from the parquet file

    """
    for i in range(vulns_id_array.shape[0]):
        vuln_dict[vulns_id_array[i]] = np.int32(i)


@nb.njit(cache=True)
def update_vuln_array_with_adj_data(vuln_array, vuln_dict, adj_vuln_data):
    """
    Update the vulnerability array with adjustment data (used for parquet loading).

    Args:
        vuln_array: (3D array) The vulnerability data array.
        vuln_dict: (Dict[int, int]) Maps vulnerability IDs to indices in vuln_array.
        adj_vuln_data: (List[vulnerability_dtype]) The vulnerability adjustment data.

    Returns: (3D array) The updated vulnerability data array.
    """
    for adj_vuln in adj_vuln_data:
        vuln_id = adj_vuln['vulnerability_id']
        if vuln_id in vuln_dict:
            load_vuln_probability(vuln_array[vuln_dict[vuln_id]], adj_vuln, vuln_id)
    return vuln_array


@nb.njit(cache=True)
def create_vulns_id(vuln_dict):
    """
    Creates a vulnerability array where the index of the array correlates with the index of the vulnerability.

    Args:
        vuln_dict: (Dict) maps the vulnerability of the id (key) with the vulnerability ID (value)

    Returns: (List[int]) list of vulnerability IDs
    """
    vulns_id = np.empty(len(vuln_dict), dtype=np.int32)

    for vuln_id, vuln_idx in vuln_dict.items():
        vulns_id[vuln_idx] = vuln_id

    return vulns_id


def get_vulns(
        storage: BaseStorage, run_dir, vuln_dict, num_intensity_bins, ignore_file_type=set(), df_engine="oasis_data_manager.df_reader.reader.OasisPandasReader"):
    """
    Loads the vulnerabilities from the file.

    Args:
        storage: (str) the storage manager for fetching model data
        run_dir: (str) the path to the run folder (used to load the analysis settings)
        vuln_dict: (Dict[int, int]) maps the vulnerability ID with the index in the vulnerability array
        num_intensity_bins: (int) the number of intensity bins
        ignore_file_type: set(str) file extension to ignore when loading

    Returns: (Tuple[List[List[float]], int, np.array[int]) vulnerability data, vulnerabilities id, number of damage bins
    """
    input_files = set(storage.listdir())
    vuln_adj = get_vulnerability_replacements(run_dir, vuln_dict)

    if vulnerability_dataset in input_files and "parquet" not in ignore_file_type:
        source_url = storage.get_storage_url(vulnerability_dataset, encode_params=False)[1]
        with storage.open(parquetvulnerability_meta_filename, 'r') as outfile:
            meta_data = json.load(outfile)
        logger.debug(f"loading {source_url}")

        df_reader_config = clean_config(InputReaderConfig(filepath=vulnerability_dataset, engine=df_engine))
        df_reader_config["engine"]["options"]["storage"] = storage
        reader = get_df_reader(df_reader_config, filters=[[('vulnerability_id', '==', vuln_id)] for vuln_id in vuln_dict.keys()])
        df = reader.as_pandas()
        num_damage_bins = meta_data['num_damage_bins']
        vuln_array = np.vstack(df['vuln_array'].to_numpy()).reshape(len(df['vuln_array']),
                                                                    num_damage_bins,
                                                                    num_intensity_bins)
        vuln_ids = df['vulnerability_id'].to_numpy()
        missing_vuln_ids = set(vuln_dict).difference(vuln_ids)
        if missing_vuln_ids:
            raise Exception(f"Vulnerability_ids {missing_vuln_ids} are missing"
                            f" from {source_url}")
        update_vulns_dictionary(vuln_dict, vuln_ids)
        # update vulnerability array with adjustment data if present
        if vuln_adj is not None and len(vuln_adj) > 0:
            vuln_array = update_vuln_array_with_adj_data(vuln_array, vuln_dict, vuln_adj)

    else:
        if "vulnerability.bin" in input_files and 'bin' not in ignore_file_type:
            source_url = storage.get_storage_url('vulnerability.bin', encode_params=False)[1]
            logger.debug(f"loading {source_url}")
            with storage.open("vulnerability.bin", 'rb') as f:
                header = np.frombuffer(f.read(8), 'i4')
                num_damage_bins = header[0]

            if "vulnerability.idx" in input_files and 'idx' not in ignore_file_type:
                logger.debug(f"loading {storage.get_storage_url('vulnerability.idx', encode_params=False)[1]}")
                with storage.open("vulnerability.bin") as f:
                    vulns_bin = np.memmap(f, dtype=VulnerabilityRow, offset=4, mode='r')

                with storage.open("vulnerability.idx") as f:
                    vulns_idx_bin = np.memmap(f, dtype=VulnerabilityIndex, mode='r')

                if vuln_adj is not None and len(vuln_adj) > 0:
                    vuln_array, valid_vuln_ids = load_vulns_bin_idx_adjusted(vulns_bin, vulns_idx_bin, vuln_dict,
                                                                             num_damage_bins, num_intensity_bins, VulnerabilityRow.dtype.itemsize, vuln_adj)
                else:
                    vuln_array, valid_vuln_ids = load_vulns_bin_idx(vulns_bin, vulns_idx_bin, vuln_dict,
                                                                    num_damage_bins, num_intensity_bins, VulnerabilityRow.dtype.itemsize)
            else:
                with storage.with_fileno("vulnerability.bin") as f:
                    vulns_bin = np.memmap(f, dtype=vulnerability_dtype, offset=4, mode='r')
                if vuln_adj is not None and len(vuln_adj) > 0:
                    vuln_array, valid_vuln_ids = load_vulns_bin_adjusted(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins, vuln_adj)
                else:
                    vuln_array, valid_vuln_ids = load_vulns_bin(vulns_bin, vuln_dict, num_damage_bins, num_intensity_bins)

        elif "vulnerability.csv" in input_files and "csv" not in ignore_file_type:
            source_url = storage.get_storage_url('vulnerability.csv', encode_params=False)[1]
            logger.debug(f"loading {source_url}")
            with storage.open("vulnerability.csv") as f:
                vuln_csv = np.loadtxt(f, dtype=vulnerability_dtype, delimiter=",", skiprows=1, ndmin=1)
            num_damage_bins = max(vuln_csv['damage_bin_id'])
            if vuln_adj is not None and len(vuln_adj) > 0:
                vuln_array, valid_vuln_ids = load_vulns_bin_adjusted(vuln_csv, vuln_dict, num_damage_bins, num_intensity_bins, vuln_adj)
            else:
                vuln_array, valid_vuln_ids = load_vulns_bin(vuln_csv, vuln_dict, num_damage_bins, num_intensity_bins)
        else:
            raise FileNotFoundError(f"vulnerability file not found at {storage.get_storage_url('', encode_params=False)[1]}")
        missing_vuln_ids = set(vuln_dict).difference(valid_vuln_ids)
        if missing_vuln_ids:
            raise Exception(f"Vulnerability_ids {missing_vuln_ids} are missing"
                            f" from {source_url}")
        vuln_ids = create_vulns_id(vuln_dict)

    return vuln_array, vuln_ids, num_damage_bins


def get_vulnerability_replacements(run_dir, vuln_dict):
    """
    Loads the vulnerability adjustment file.

    Args:
        path: (str) the path pointing to the run directory
        vuln_dict: (Dict[int, int]) list of vulnerability IDs

    Returns: (List[vulnerability_dtype]) vulnerability replacement data
    """
    settings_path = os.path.join(run_dir, "analysis_settings.json")

    if not os.path.exists(settings_path):
        logger.warning(f"analysis_settings.json not found in {run_dir}.")
        return None

    if not validate_vulnerability_replacements(settings_path):
        return None

    vulnerability_replacements_key = analysis_settings_loader(settings_path).get('vulnerability_adjustments')

    # Inputting the data directly into the analysis_settings.json file takes precedence over the file path
    vulnerability_replacements_field = vulnerability_replacements_key.get('replace_data', None)
    if vulnerability_replacements_field is None:
        vulnerability_replacements_field = vulnerability_replacements_key.get('replace_file', None)

    if isinstance(vulnerability_replacements_field, dict):
        # Convert dict to flat array equivalent to csv format
        flat_data = []
        for v_id, adjustments in vulnerability_replacements_field.items():
            for adj in adjustments:
                flat_data.append((v_id, *adj))
        vuln_adj = np.array(flat_data, dtype=vulnerability_dtype)
    elif isinstance(vulnerability_replacements_field, str):
        # Load csv file
        absolute_path = os.path.abspath(vulnerability_replacements_field)
        logger.debug(f"loading {absolute_path}")
        vuln_adj = np.loadtxt(absolute_path, dtype=vulnerability_dtype, delimiter=",", skiprows=1, ndmin=1)
    vuln_adj = np.array([adj_vuln for adj_vuln in vuln_adj if adj_vuln['vulnerability_id'] in vuln_dict],
                        dtype=vuln_adj.dtype)
    vuln_adj.sort(order='vulnerability_id')
    logger.info("Vulnerability adjustments found in analysis settings.")
    return vuln_adj


def get_mean_damage_bins(storage: BaseStorage, ignore_file_type=set()):
    """
    Loads the mean damage bins from the damage_bin_dict file, namely, the `interpolation` value for each bin.

    Args:
        storage: (BaseStorage) the storage connector for fetching the model data
        ignore_file_type: set(str) file extension to ignore when loading

    Returns: (List[Union[damagebindictionary]]) loaded data from the damage_bin_dict file
    """
    return get_damage_bins(storage, ignore_file_type)['interpolation']


def get_damage_bins(storage: BaseStorage, ignore_file_type=set()):
    """
    Loads the damage bins from the damage_bin_dict file.

    Args:
        storage: (BaseStorage) the storage connector for fetching the model data
        ignore_file_type: set(str) file extension to ignore when loading

    Returns: (List[Union[damagebindictionary]]) loaded data from the damage_bin_dict file
    """
    input_files = set(storage.listdir())

    if "damage_bin_dict.bin" in input_files and 'bin' not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('damage_bin_dict.bin', encode_params=False)[1]}")
        with storage.with_fileno("damage_bin_dict.bin") as f:
            return np.fromfile(f, dtype=damagebin_dtype)
    elif "damage_bin_dict.csv" in input_files and 'csv' not in ignore_file_type:
        logger.debug(f"loading {storage.get_storage_url('damage_bin_dict.csv', encode_params=False)[1]}")
        with storage.open("damage_bin_dict.csv") as f:
            return np.loadtxt(f, dtype=damagebin_dtype, skiprows=1, delimiter=',', ndmin=1)
    else:
        raise FileNotFoundError(f"damage_bin_dict file not found at {storage.get_storage_url('', encode_params=False)[1]}")


@nb.njit(cache=True, fastmath=True)
def damage_bin_prob(p, intensities_min, intensities_max, vulns, intensities):
    """
    Calculate the probability of an event happening and then causing damage.
    Note: vulns is a 1-d array containing 1 damage bin of the damage probability distribution as a
    function of hazard intensity.

    Args:
        p: (float) the probability to be updated
        intensities_min: (int) minimum intensity bin id
        intensities_max: (int) maximum intensity bin id
        vulns: (List[float]) slice of damage probability distribution given hazard intensity
        intensities: (List[float]) intensity probability distribution

    Returns: (float) the updated probability
    """
    i = intensities_min
    while i < intensities_max:
        p += vulns[i] * intensities[i]
        i += 1
    return p


@nb.njit(cache=True, fastmath=True)
def do_result(vulns_id, vuln_array, mean_damage_bins,
              int32_mv, num_damage_bins,
              intensities_min, intensities_max, intensities,
              event_id, areaperil_id, vuln_i, cursor):
    """
    Calculate the result concerning an event ID.

    Args:
        vulns_id: (List[int]) list of vulnerability IDs
        vuln_array: (List[List[list]]) list of vulnerabilities and their data
        mean_damage_bins: (List[float]) the mean of each damage bin (len(mean_damage_bins) == num_damage_bins)
        int32_mv: (List[int]) FILL IN LATER
        num_damage_bins: (int) number of damage bins in the data
        intensities_min: (int) minimum intensity bin id
        intensities_max: (int) maximum intensity bin id
        intensities: (List[float]) intensity probability distribution
        event_id: (int) the event ID that concerns the result being calculated
        areaperil_id: (List[int]) the areaperil ID that concerns the result being calculated
        vuln_i: (int) the index concerning the vulnerability inside the vuln_array
        cursor: (int) PLEASE FILL IN

    Returns: (int) PLEASE FILL IN
    """
    int32_mv[cursor], cursor = event_id, cursor + 1
    int32_mv[cursor:cursor + areaperil_int_relative_size] = areaperil_id.view(oasis_int)
    cursor += areaperil_int_relative_size
    int32_mv[cursor], cursor = vulns_id[vuln_i], cursor + 1

    cur_vuln_mat = vuln_array[vuln_i]
    p = 0
    cursor_start = cursor
    cursor += 1
    oasis_float_mv = int32_mv[cursor: cursor + num_damage_bins * results_relative_size].view(oasis_float)
    result_cursor = 0
    damage_bin_i = 0

    while damage_bin_i < num_damage_bins:
        p = damage_bin_prob(p, intensities_min, intensities_max, cur_vuln_mat[damage_bin_i], intensities)
        oasis_float_mv[result_cursor], result_cursor = p, result_cursor + 1
        oasis_float_mv[result_cursor], result_cursor = mean_damage_bins[damage_bin_i], result_cursor + 1
        damage_bin_i += 1
        if p >= 0.999999940:
            break

    int32_mv[cursor_start] = damage_bin_i
    return cursor + (result_cursor * oasis_float_relative_size)


@nb.njit(cache=True)
def doCdf(event_id,
          num_intensity_bins, footprint,
          areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns,
          vuln_array, vulns_id, num_damage_bins, mean_damage_bins,
          int32_mv, max_result_relative_size):
    """
    Calculates the cumulative distribution function (cdf) for an event ID.

    Args:
        event_id: (int) the event ID the the CDF is being calculated to.
        num_intensity_bins: (int) the number of intensity bins for the CDF
        footprint: (List[Tuple[int, int, float]]) information about the footprint with event_id, areaperil_id,
                                                  probability
        areaperil_to_vulns_idx_dict: (Dict[int, int]) maps the areaperil ID with the ENTER_HERE
        areaperil_to_vulns_idx_array: (List[Tuple[int, int]]) the index where the areaperil ID starts and finishes
        areaperil_to_vulns: (List[int]) maps the areaperil ID to the vulnerability ID
        vuln_array: (List[list]) FILL IN LATER
        vulns_id: (List[int]) list of vulnerability IDs
        num_damage_bins: (int) number of damage bins in the data
        mean_damage_bins: (List[float]) the mean of each damage bin (len(mean_damage_bins) == num_damage_bins)
        int32_mv: (List[int]) FILL IN LATER
        max_result_relative_size: (int) the maximum result size

    Returns: (int)
    """
    if not footprint.shape[0]:
        return 0

    intensities_min = num_intensity_bins
    intensities_max = 0
    intensities = np.zeros(num_intensity_bins, dtype=oasis_float)

    areaperil_id = np.zeros(1, dtype=areaperil_int)
    has_vuln = False
    cursor = 0

    for footprint_i in range(footprint.shape[0]):
        event_row = footprint[footprint_i]
        if areaperil_id[0] != event_row['areaperil_id']:
            if has_vuln and intensities_min <= intensities_max:
                areaperil_to_vulns_idx = areaperil_to_vulns_idx_array[areaperil_to_vulns_idx_dict[areaperil_id[0]]]
                intensities_max += 1
                for vuln_idx in range(areaperil_to_vulns_idx['start'], areaperil_to_vulns_idx['end']):
                    vuln_i = areaperil_to_vulns[vuln_idx]
                    if cursor + max_result_relative_size > buff_int_size:
                        yield cursor * oasis_int_size
                        cursor = 0

                    cursor = do_result(vulns_id, vuln_array, mean_damage_bins,
                                       int32_mv, num_damage_bins,
                                       intensities_min, intensities_max, intensities,
                                       event_id, areaperil_id, vuln_i, cursor)

            areaperil_id[0] = event_row['areaperil_id']
            has_vuln = areaperil_id[0] in areaperil_to_vulns_idx_dict

            if has_vuln:
                intensities[intensities_min: intensities_max] = 0
                intensities_min = num_intensity_bins
                intensities_max = 0
        if has_vuln:
            if event_row['probability'] > 0:
                intensity_bin_i = event_row['intensity_bin_id'] - 1
                intensities[intensity_bin_i] = event_row['probability']
                if intensity_bin_i > intensities_max:
                    intensities_max = intensity_bin_i
                if intensity_bin_i < intensities_min:
                    intensities_min = intensity_bin_i

    if has_vuln and intensities_min <= intensities_max:
        areaperil_to_vulns_idx = areaperil_to_vulns_idx_array[areaperil_to_vulns_idx_dict[areaperil_id[0]]]
        intensities_max += 1
        for vuln_idx in range(areaperil_to_vulns_idx['start'], areaperil_to_vulns_idx['end']):
            vuln_i = areaperil_to_vulns[vuln_idx]
            if cursor + max_result_relative_size > buff_int_size:
                yield cursor * oasis_int_size
                cursor = 0

            cursor = do_result(vulns_id, vuln_array, mean_damage_bins,
                               int32_mv, num_damage_bins,
                               intensities_min, intensities_max, intensities,
                               event_id, areaperil_id, vuln_i, cursor)

    yield cursor * oasis_int_size


@nb.njit()
def convert_vuln_id_to_index(vuln_dict, areaperil_to_vulns):
    for i in range(areaperil_to_vulns.shape[0]):
        areaperil_to_vulns[i] = vuln_dict[areaperil_to_vulns[i]]


@redirect_logging(exec_name='modelpy')
def run(
    run_dir,
    file_in,
    file_out,
    ignore_file_type,
    data_server,
    peril_filter,
    df_engine="oasis_data_manager.df_reader.reader.OasisPandasReader",
    analysis_pk=None
):
    """
    Runs the main process of the getmodel process.

    Args:
        run_dir: (str) the directory of where the process is running
        file_in: (Optional[str]) the path to the input directory
        file_out: (Optional[str]) the path to the output directory
        ignore_file_type: set(str) file extension to ignore when loading
        data_server: (bool) if set to True runs the data server
        peril_filter (list[int]): list of perils to include in the computation (if None, all perils will be included).
        df_engine: (str) The engine to use when loading dataframes

    Returns: None
    """
    model_storage = get_storage_from_config_path(
        os.path.join(run_dir, 'model_storage.json'),
        os.path.join(run_dir, 'static'),
    )
    input_path = os.path.join(run_dir, 'input')
    ignore_file_type = set(ignore_file_type)

    if data_server:
        logger.debug("data server active")
        FootprintLayerClient.register()
        logger.debug("registered with data server")
        atexit.register(FootprintLayerClient.unregister)
    else:
        logger.debug("data server not active")

    with ExitStack() as stack:
        if file_in is None:
            streams_in = sys.stdin.buffer
        else:
            streams_in = stack.enter_context(open(file_in, 'rb'))

        if file_out is None:
            stream_out = sys.stdout.buffer
        else:
            stream_out = stack.enter_context(open(file_out, 'wb'))

        event_id_mv = memoryview(bytearray(4))
        event_ids = np.ndarray(1, buffer=event_id_mv, dtype='i4')

        # load keys.csv to determine included AreaPerilID from peril_filter
        if os.path.exists(os.path.join(input_path, 'keys.csv')):
            keys_df = pd.read_csv(os.path.join(input_path, 'keys.csv'), dtype=Keys)
            if peril_filter:
                valid_area_peril_id = np.unique(keys_df.loc[keys_df['PerilID'].isin(peril_filter), 'AreaPerilID'])
                logger.debug(
                    f'Peril specific run: ({peril_filter}), {len(valid_area_peril_id)} AreaPerilID included out of {len(keys_df)}')
            else:
                valid_area_peril_id = keys_df['AreaPerilID']
        else:
            valid_area_peril_id = None

        logger.debug('init items')
        vuln_dict, areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns = get_items(
            input_path, ignore_file_type, valid_area_peril_id if peril_filter else None)

        logger.debug('init footprint')
        footprint_obj = stack.enter_context(
            Footprint.load(model_storage, ignore_file_type, df_engine=df_engine,
                           areaperil_ids=list(areaperil_to_vulns_idx_dict.keys())))

        if data_server:
            num_intensity_bins: int = FootprintLayerClient.get_number_of_intensity_bins()
            logger.info(f"got {num_intensity_bins} intensity bins from server")
        else:
            num_intensity_bins: int = footprint_obj.num_intensity_bins

        logger.debug('init vulnerability')

        vuln_array, vulns_id, num_damage_bins = get_vulns(model_storage, run_dir, vuln_dict,
                                                          num_intensity_bins, ignore_file_type, df_engine=df_engine)
        convert_vuln_id_to_index(vuln_dict, areaperil_to_vulns)
        logger.debug('init mean_damage_bins')
        mean_damage_bins = get_mean_damage_bins(model_storage, ignore_file_type)

        # even_id, areaperil_id, vulnerability_id, num_result, [oasis_float] * num_result
        max_result_relative_size = 1 + + areaperil_int_relative_size + 1 + 1 + num_damage_bins * results_relative_size
        mv = memoryview(bytearray(buff_size))
        int32_mv = np.ndarray(buff_size // np.int32().itemsize, buffer=mv, dtype=np.int32)

        # header
        stream_out.write(np.uint32(1).tobytes())

        logger.debug('doCdf starting')
        empty_events = 0
        while True:
            len_read = streams_in.readinto(event_id_mv)
            if len_read == 0:
                break

            # get the next event_id from the input stream
            event_id = event_ids[0]

            if data_server:
                event_footprint = FootprintLayerClient.get_event(event_id)
            else:
                event_footprint = footprint_obj.get_event(event_id)

            if event_footprint is not None:
                # compute effective damageability probability distribution
                # stream out: event_id, areaperil_id, number of damage bins, effecive damageability cdf bins (bin_mean and prob_to)
                for cursor_bytes in doCdf(event_id,
                                          num_intensity_bins, event_footprint,
                                          areaperil_to_vulns_idx_dict, areaperil_to_vulns_idx_array, areaperil_to_vulns,
                                          vuln_array, vulns_id, num_damage_bins, mean_damage_bins,
                                          int32_mv, max_result_relative_size):

                    if cursor_bytes:
                        stream_out.write(mv[:cursor_bytes])
                    else:
                        empty_events += 1
                        break
            else:
                empty_events += 1
        oasis_ping({"events_complete": empty_events, 'analysis_pk': analysis_pk})



import pandas as pd
import numpy as np
from oasislmf.lookup.builtin import normal_to_z_index
from oasislmf.utils.status import OASIS_UNKNOWN_ID


def change_footprint_apid(path, size_lat):
    """Reads CSV, applies Z-index conversion, and writes back."""
    change_footprint_apid_multi_peril(path, size_lat, None, 1)


def change_footprint_apid_multi_peril(path, size_lat, size_lon, num_perils):
    """
    Reads CSV, applies Z-index conversion for multiple perils and writes
    back to file
    """
    df = pd.read_csv(path)

    areaperil_ids = df['areaperil_id'].values - 1

    mask = df['areaperil_id'] != OASIS_UNKNOWN_ID
    filtered_df = df[mask].copy()

    areaperil_ids = filtered_df['areaperil_id'].values - 1

    if size_lon is not None:
        grid_size = size_lat * size_lon
        peril_values = areaperil_ids // grid_size
        index_values = areaperil_ids % grid_size + 1
    else:
        peril_values = areaperil_ids // 1
        index_values = areaperil_ids + 1

    z_indices = np.array(
        [normal_to_z_index(id, size_lat) for id in index_values]
    )

    filtered_df['areaperil_id'] = z_indices * num_perils + peril_values
    df.loc[mask, 'areaperil_id'] = filtered_df['areaperil_id']

    df = df.sort_values(by=df.columns.tolist(), kind='stable')
    df.to_csv(path, index=False)



__all__ = [
    'GENERAL_SETTINGS_FILE',
    'GUL_INPUT_FILES',
    'GUL_SUMMARIES_FILE',
    'INPUT_FILES',
    'IL_INPUT_FILES',
    'IL_SUMMARIES_FILE',
    'MODEL_SETTINGS_FILE',
    'OPTIONAL_INPUT_FILES',
    'TAR_FILE'
]

INPUT_FILES = {
    'complex_items': {
        'name': 'complex_items',
        'type': 'gul',
        'conversion_tool': 'complex_itemtobin',
        'csvtobin_type': 'complex_items',
    },
    'items': {
        'name': 'items',
        'type': 'gul',
        'conversion_tool': 'itemtobin',
        'csvtobin_type': 'items',
    },
    'coverages': {
        'name': 'coverages',
        'type': 'gul',
        'conversion_tool': 'coveragetobin',
        'csvtobin_type': 'coverages',
    },
    'gulsummaryxref': {
        'name': 'gulsummaryxref',
        'type': 'gul',
        'conversion_tool': 'gulsummaryxreftobin',
        'csvtobin_type': 'gul_summary_xref',
    },
    'events': {
        'name': 'events',
        'type': 'optional',
        'conversion_tool': 'evetobin',
        'csvtobin_type': 'eve',
    },
    'amplifications': {
        'name': 'amplifications',
        'type': 'optional',
        'conversion_tool': 'amplificationstobin',
        'csvtobin_type': 'amplifications',
    },
    'fm_policytc': {
        'name': 'fm_policytc',
        'type': 'il',
        'conversion_tool': 'fmpolicytctobin',
        'csvtobin_type': 'fm_policytc',
    },
    'fm_profile': {
        'name': 'fm_profile',
        'type': 'il',
        'conversion_tool': 'fmprofiletobin',
        'step_flag': '-S',
        'csvtobin_type': 'fm_profile',
    },
    'fm_programme': {
        'name': 'fm_programme',
        'type': 'il',
        'conversion_tool': 'fmprogrammetobin',
        'csvtobin_type': 'fm_programme',
    },
    'fm_xref': {
        'name': 'fm_xref',
        'type': 'il',
        'conversion_tool': 'fmxreftobin',
        'csvtobin_type': 'fm_xref',
    },
    'fmsummaryxref': {
        'name': 'fmsummaryxref',
        'type': 'il',
        'conversion_tool': 'fmsummaryxreftobin',
        'csvtobin_type': 'fm_summary_xref',
    }
}
GUL_INPUT_FILES = {k: v for k, v in INPUT_FILES.items() if v['type'] == 'gul'}
IL_INPUT_FILES = {k: v for k, v in INPUT_FILES.items() if v['type'] == 'il'}
OPTIONAL_INPUT_FILES = {k: v for k, v in INPUT_FILES.items() if v['type'] == 'optional'}

TAR_FILE = 'inputs.tar.gz'

GENERAL_SETTINGS_FILE = "general_settings.csv"
MODEL_SETTINGS_FILE = "model_settings.csv"
GUL_SUMMARIES_FILE = "gul_summaries.csv"
IL_SUMMARIES_FILE = "il_summaries.csv"



#!/usr/bin/env python

import argparse
import logging

from . import manager, logger


def main():
    parser = argparse.ArgumentParser(description='Join Summary ID info to data file and output new file')
    parser.add_argument('-s', '--summaryinfo', type=str, required=True, default=None, help='Summary Info File')
    parser.add_argument('-d', '--data', type=str, required=True, default=None, help='Ouput Data File')
    parser.add_argument('-o', '--output', type=str, required=True, default=None, help='Joined Output File')
    parser.add_argument('-v', '--logging-level', type=int, default=30,
                        help='logging level (debug:10, info:20, warning:30, error:40, critical:50)')

    args = parser.parse_args()
    kwargs = vars(args)

    # Set up logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.main(**kwargs)


if __name__ == '__main__':
    main()



import sys
import zlib
import numpy as np
from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, resolve_file, write_ndarray_to_fmt_csv
from oasislmf.pytools.converters.data import TOOL_INFO
from oasislmf.pytools.getmodel.manager import VulnerabilityIndex, VulnerabilityRow_dtype
from oasislmf.utils.exceptions import OasisException


def _read_bin_idx_zip(file_in, idx_file_in, dtype):
    if file_in == sys.stdin.buffer:
        bin_data = np.frombuffer(file_in.read(), dtype="u1")
    else:
        bin_data = np.fromfile(file_in, dtype="u1")

    idx_data = np.memmap(idx_file_in, dtype=VulnerabilityIndex)
    num_rows = np.sum(idx_data["original_size"]) // VulnerabilityRow_dtype.itemsize

    data = np.empty(num_rows, dtype=dtype)
    data_index = 0
    for row in idx_data:
        vuln_id = row["vulnerability_id"]
        offset = row["offset"]
        size = row["size"]
        original_size = row["original_size"]
        num_data = original_size // VulnerabilityRow_dtype.itemsize
        zipped_data = bin_data[offset:offset + size]
        unzipped_data = zlib.decompress(zipped_data)
        vuln_row_data = np.frombuffer(unzipped_data, dtype=VulnerabilityRow_dtype)
        data[data_index:data_index + num_data]["vulnerability_id"] = vuln_id
        data[data_index:data_index + num_data]["intensity_bin_id"] = vuln_row_data["intensity_bin_id"]
        data[data_index:data_index + num_data]["damage_bin_id"] = vuln_row_data["damage_bin_id"]
        data[data_index:data_index + num_data]["probability"] = vuln_row_data["probability"]
        data_index += num_data
    return data


def _read_bin_idx(file_in, idx_file_in, dtype):
    if file_in == sys.stdin.buffer:
        bin_data = np.frombuffer(file_in.read(), dtype="u1")
    else:
        bin_data = np.fromfile(file_in, dtype="u1")

    idx_data = np.memmap(idx_file_in, dtype=VulnerabilityIndex)
    num_rows = np.sum(idx_data["size"]) // VulnerabilityRow_dtype.itemsize

    data = np.empty(num_rows, dtype=dtype)
    data_index = 0
    for row in idx_data:
        vuln_id = row["vulnerability_id"]
        offset = row["offset"]
        size = row["size"]
        num_data = size // VulnerabilityRow_dtype.itemsize
        vuln_row_data = bin_data[offset:offset + size].view(dtype=VulnerabilityRow_dtype)
        data[data_index:data_index + num_data]["vulnerability_id"] = vuln_id
        data[data_index:data_index + num_data]["intensity_bin_id"] = vuln_row_data["intensity_bin_id"]
        data[data_index:data_index + num_data]["damage_bin_id"] = vuln_row_data["damage_bin_id"]
        data[data_index:data_index + num_data]["probability"] = vuln_row_data["probability"]
        data_index += num_data
    return data


def _read_bin(file_in, dtype):
    # Offet 4 bytes for damage bins int32
    if file_in == sys.stdin.buffer:
        return np.frombuffer(file_in.read(), dtype=dtype, offset=4)
    else:
        return np.fromfile(file_in, dtype=dtype, offset=4)


def vulnerability_tocsv(stack, file_in, file_out, file_type, noheader, idx_file_in, zip_files):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    file_in = resolve_file(file_in, "rb", stack)

    if not noheader:
        file_out.write(",".join(headers) + "\n")

    if zip_files:
        if idx_file_in is None:
            raise OasisException(f"Error: Cannot read zip files without provided idx_file_in zip path, currently {idx_file_in}")
        data = _read_bin_idx_zip(file_in, idx_file_in, dtype)
    else:
        if idx_file_in is not None:
            data = _read_bin_idx(file_in, idx_file_in, dtype)
        else:
            data = _read_bin(file_in, dtype)

    num_rows = data.shape[0]
    buffer_size = DEFAULT_BUFFER_SIZE
    for start in range(0, num_rows, buffer_size):
        end = min(start + buffer_size, num_rows)
        buffer_data = data[start:end]
        write_ndarray_to_fmt_csv(file_out, buffer_data, headers, fmt)



from pathlib import Path
import numpy as np
from oasislmf.pytools.common.data import write_ndarray_to_fmt_csv
from oasislmf.pytools.common.input_files import read_amplifications
from oasislmf.pytools.converters.data import TOOL_INFO


def amplifications_tocsv(stack, file_in, file_out, file_type, noheader):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    fmt = TOOL_INFO[file_type]["fmt"]

    if str(file_in) == "-":
        items_amps = read_amplifications(use_stdin=True)
    else:
        amps_fp = Path(file_in)
        items_amps = read_amplifications(run_dir=amps_fp.parent, filename=amps_fp.name)
    items_amps = items_amps[1:]
    data = np.zeros(len(items_amps), dtype=dtype)
    data["item_id"] = np.arange(1, len(items_amps) + 1)
    data["amplification_id"] = items_amps

    if not noheader:
        file_out.write(",".join(headers) + "\n")
    write_ndarray_to_fmt_csv(file_out, data, headers, fmt)



import os
import sys
import numba as nb
import numpy as np
import pandas as pd


oasis_int = np.dtype(os.environ.get('OASIS_INT', 'i4'))
nb_oasis_int = nb.from_dtype(oasis_int)
oasis_int_size = oasis_int.itemsize

oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))
nb_oasis_float = nb.from_dtype(oasis_float)
oasis_float_size = oasis_float.itemsize

areaperil_int = np.dtype(os.environ.get('AREAPERIL_TYPE', 'u4'))
nb_areaperil_int = nb.from_dtype(areaperil_int)
areaperil_int_size = areaperil_int.itemsize

null_index = oasis_int.type(-1)

# A default buffer size for nd arrays to be initialised to
DEFAULT_BUFFER_SIZE = 1_000_000

# Mean type numbers for outputs (SampleType)
MEAN_TYPE_ANALYTICAL = 1
MEAN_TYPE_SAMPLE = 2


def generate_output_metadata(output):
    """Generates *_header, *_dtype and *_fmt items given a list of tuples describing some output description
    output description has type List(Tuple({name: str}, {type: Any}, {format: str}))
    Args:
        output_map (list(tuple(str, Any, str))): Dictionary mapping string name to  {output description}_output list
    Returns:
        result (tuple(list[str], np.dtype, str)): Tuple containing the generated *_header list, *_dtype np.dtype, *_fmt csv format string
    """
    headers = [c[0] for c in output]
    dtype = np.dtype([(c[0], c[1]) for c in output])
    fmt = ','.join([c[2] for c in output])
    result = (headers, dtype, fmt)
    return result


# Types
aggregatevulnerability_output = [
    ("aggregate_vulnerability_id", 'i4', "%d"),
    ("vulnerability_id", 'i4', "%d"),
]
aggregatevulnerability_headers, aggregatevulnerability_dtype, aggregatevulnerability_fmt = generate_output_metadata(aggregatevulnerability_output)

amplifications_output = [
    ("item_id", 'i4', "%d"),
    ("amplification_id", 'i4', "%d"),
]
amplifications_headers, amplifications_dtype, amplifications_fmt = generate_output_metadata(amplifications_output)

cdf_output = [
    ("event_id", 'i4', "%d"),
    ("areaperil_id", 'i4', "%d"),
    ("vulnerability_id", 'i4', "%d"),
    ("bin_index", 'i4', "%d"),
    ("prob_to", 'f4', "%f"),
    ("bin_mean", 'f4', "%f"),
]
cdf_headers, cdf_dtype, cdf_fmt = generate_output_metadata(cdf_output)

complex_items_meta_output = [
    ("item_id", 'u4', "%u"),
    ("coverage_id", 'u4', "%u"),
    ("group_id", 'u4', "%u"),
    ("model_data_len", 'u4', "%u"),
]
complex_items_meta_headers, complex_items_meta_dtype, complex_items_meta_fmt = generate_output_metadata(complex_items_meta_output)

correlations_output = [
    ("item_id", 'i4', "%d"),
    ("peril_correlation_group", 'i4', "%d"),
    ("damage_correlation_value", oasis_float, "%f"),
    ("hazard_group_id", 'i4', "%d"),
    ("hazard_correlation_value", oasis_float, "%f"),
]
correlations_headers, correlations_dtype, correlations_fmt = generate_output_metadata(correlations_output)

coverages_output = [
    ("coverage_id", 'i4', "%d"),
    ("tiv", oasis_float, "%f"),
]
coverages_headers, coverages_dtype, coverages_fmt = generate_output_metadata(coverages_output)

damagebin_output = [
    ("bin_index", 'i4', "%d"),
    ("bin_from", oasis_float, "%f"),
    ("bin_to", oasis_float, "%f"),
    ("interpolation", oasis_float, "%f"),
    ("damage_type", 'i4', "%d"),
]
damagebin_headers, damagebin_dtype, damagebin_fmt = generate_output_metadata(damagebin_output)

eve_output = [
    ("event_id", oasis_int, "%d")
]
eve_headers, eve_dtype, eve_fmt = generate_output_metadata(eve_output)

footprint_event_output = [
    ('event_id', 'i4', "%d"),
    ('areaperil_id', areaperil_int, "%d"),
    ('intensity_bin_id', 'i4', "%d"),
    ('probability', oasis_float, "%.6f"),
]
footprint_event_headers, footprint_event_dtype, footprint_event_fmt = generate_output_metadata(footprint_event_output)

fm_output = [
    ("event_id", 'i4', "%d"),
    ("output_id", 'i4', "%d"),
    ("sidx", 'i4', "%d"),
    ("loss", oasis_float, "%.2f"),
]
fm_headers, fm_dtype, fm_fmt = generate_output_metadata(fm_output)

fm_policytc_output = [
    ("level_id", 'i4', "%d"),
    ("agg_id", 'i4', "%d"),
    ("layer_id", 'i4', "%d"),
    ("profile_id", 'i4', "%d"),
]
fm_policytc_headers, fm_policytc_dtype, fm_policytc_fmt = generate_output_metadata(fm_policytc_output)

fm_profile_output = [
    ("profile_id", 'i4', "%d"),
    ("calcrule_id", 'i4', "%d"),
    ("deductible1", oasis_float, "%f"),
    ("deductible2", oasis_float, "%f"),
    ("deductible3", oasis_float, "%f"),
    ("attachment1", oasis_float, "%f"),
    ("limit1", oasis_float, "%f"),
    ("share1", oasis_float, "%f"),
    ("share2", oasis_float, "%f"),
    ("share3", oasis_float, "%f"),
]
fm_profile_headers, fm_profile_dtype, fm_profile_fmt = generate_output_metadata(fm_profile_output)

fm_profile_step_output = [
    ("profile_id", 'i4', "%d"),
    ("calcrule_id", 'i4', "%d"),
    ("deductible1", oasis_float, "%f"),
    ("deductible2", oasis_float, "%f"),
    ("deductible3", oasis_float, "%f"),
    ("attachment1", oasis_float, "%f"),
    ("limit1", oasis_float, "%f"),
    ("share1", oasis_float, "%f"),
    ("share2", oasis_float, "%f"),
    ("share3", oasis_float, "%f"),
    ("step_id", 'i4', "%d"),
    ("trigger_start", oasis_float, "%f"),
    ("trigger_end", oasis_float, "%f"),
    ("payout_start", oasis_float, "%f"),
    ("payout_end", oasis_float, "%f"),
    ("limit2", oasis_float, "%f"),
    ("scale1", oasis_float, "%f"),
    ("scale2", oasis_float, "%f"),
]
fm_profile_step_headers, fm_profile_step_dtype, fm_profile_step_fmt = generate_output_metadata(fm_profile_step_output)

fm_programme_output = [
    ("from_agg_id", 'i4', "%d"),
    ("level_id", 'i4', "%d"),
    ("to_agg_id", 'i4', "%d"),
]
fm_programme_headers, fm_programme_dtype, fm_programme_fmt = generate_output_metadata(fm_programme_output)

fm_summary_xref_output = [
    ("output", 'i4', "%d"),
    ("summary_id", 'i4', "%d"),
    ("summaryset_id", 'i4', "%d")
]
fm_summary_xref_headers, fm_summary_xref_dtype, fm_summary_xref_fmt = generate_output_metadata(fm_summary_xref_output)

fm_xref_output = [
    ("output", 'i4', "%d"),
    ("agg_id", 'i4', "%d"),
    ("layer_id", 'i4', "%d"),
]
fm_xref_headers, fm_xref_dtype, fm_xref_fmt = generate_output_metadata(fm_xref_output)

gul_output = [
    ("event_id", 'i4', "%d"),
    ("item_id", 'i4', "%d"),
    ("sidx", 'i4', "%d"),
    ("loss", oasis_float, "%.2f"),
]
gul_headers, gul_dtype, gul_fmt = generate_output_metadata(gul_output)

gul_summary_xref_output = [
    ("item_id", 'i4', "%d"),
    ("summary_id", 'i4', "%d"),
    ("summaryset_id", 'i4', "%d")
]
gul_summary_xref_headers, gul_summary_xref_dtype, gul_summary_xref_fmt = generate_output_metadata(gul_summary_xref_output)

items_output = [
    ("item_id", 'i4', "%d"),
    ("coverage_id", 'i4', "%d"),
    ("areaperil_id", areaperil_int, "%u"),
    ("vulnerability_id", 'i4', "%d"),
    ("group_id", 'i4', "%d"),
]
items_headers, items_dtype, items_fmt = generate_output_metadata(items_output)

lossfactors_output = [
    ("event_id", 'i4', "%d"),
    ("amplification_id", 'i4', "%d"),
    ("factor", 'f4', "%.2f"),
]
lossfactors_headers, lossfactors_dtype, lossfactors_fmt = generate_output_metadata(lossfactors_output)

occurrence_output = [
    ("event_id", 'i4', "%d"),
    ("period_no", 'i4', "%d"),
    ("occ_date_id", 'i4', "%d"),
]
occurrence_headers, occurrence_dtype, occurrence_fmt = generate_output_metadata(occurrence_output)

occurrence_granular_output = [
    ("event_id", 'i4', "%d"),
    ("period_no", 'i4', "%d"),
    ("occ_date_id", 'i8', "%d"),
]
occurrence_granular_headers, occurrence_granular_dtype, occurrence_granular_fmt = generate_output_metadata(occurrence_granular_output)

periods_output = [
    ("period_no", 'i4', "%d"),
    ("weighting", 'f8', "%0.9lf"),
]
periods_headers, periods_dtype, periods_fmt = generate_output_metadata(periods_output)

quantile_output = [
    ("quantile", 'f4', "%f"),
]
quantile_headers, quantile_dtype, quantile_fmt = generate_output_metadata(quantile_output)

quantile_interval_output = quantile_output + [
    ('integer_part', oasis_int, "%d"),
    ('fractional_part', oasis_float, "%f"),
]
quantile_interval_headers, quantile_interval_dtype, quantile_interval_fmt = generate_output_metadata(quantile_interval_output)

random_output = [
    ("random_no", 'f4', "%f"),
]
random_headers, random_dtype, random_fmt = generate_output_metadata(random_output)

returnperiods_output = [
    ("return_period", 'i4', "%d"),
]
returnperiods_headers, returnperiods_dtype, returnperiods_fmt = generate_output_metadata(returnperiods_output)

vulnerability_output = [
    ("vulnerability_id", 'i4', "%d"),
    ("intensity_bin_id", 'i4', "%d"),
    ("damage_bin_id", 'i4', "%d"),
    ("probability", oasis_float, "%.6f"),
]
vulnerability_headers, vulnerability_dtype, vulnerability_fmt = generate_output_metadata(vulnerability_output)

vulnerability_weight_output = [
    ("areaperil_id", areaperil_int, "%d"),
    ("vulnerability_id", 'i4', "%d"),
    ("weight", oasis_float, "%f"),
]
vulnerability_weight_headers, vulnerability_weight_dtype, vulnerability_weight_fmt = generate_output_metadata(vulnerability_weight_output)


def load_as_ndarray(dir_path, name, _dtype, must_exist=True, col_map=None):
    """
    load a file as a numpy ndarray
    useful for multi-columns files
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: np.dtype
        must_exist: raise FileNotFoundError if no file is present
        col_map: name re-mapping to change name of csv columns
    Returns:
        numpy ndarray
    """

    if os.path.isfile(os.path.join(dir_path, name + '.bin')):
        return np.fromfile(os.path.join(dir_path, name + '.bin'), dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        # in csv column cam be out of order and have different name,
        # we load with pandas and write each column to the ndarray
        if col_map is None:
            col_map = {}
        with open(os.path.join(dir_path, name + '.csv')) as file_in:
            cvs_dtype = {col_map.get(key, key): col_dtype for key, (col_dtype, _) in _dtype.fields.items()}
            df = pd.read_csv(file_in, delimiter=',', dtype=cvs_dtype, usecols=list(cvs_dtype.keys()))
            res = np.empty(df.shape[0], dtype=_dtype)
            for name in _dtype.names:
                res[name] = df[col_map.get(name, name)]
            return res
    else:
        return np.empty(0, dtype=_dtype)


def load_as_array(dir_path, name, _dtype, must_exist=True):
    """
    load file as a single numpy array,
     useful for files with a binary version with only one type of value where their index correspond to an id.
     For example coverage.bin only contains tiv value for each coverage id
     coverage_id n correspond to index n-1
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: numpy dtype of the required array
        must_exist: raise FileNotFoundError if no file is present
    Returns:
        numpy array of dtype type
    """
    fp = os.path.join(dir_path, name + '.bin')
    if os.path.isfile(fp):
        return np.fromfile(fp, dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        fp = os.path.join(dir_path, name + '.csv')
        with open(fp) as file_in:
            return np.loadtxt(file_in, dtype=_dtype, delimiter=',', skiprows=1, usecols=1)
    else:
        return np.empty(0, dtype=_dtype)


def write_ndarray_to_fmt_csv(output_file, data, headers, row_fmt):
    """Writes a custom dtype array with headers to csv with the provided row_fmt str

    This function is a faster replacement for np.savetxt as it formats each row one at a time before writing to csv.
    We create one large string, and formats all the data at once, and writes all the data at once.

    WARNING: untested with string types in custom data.

    Args:
        output_file (io.TextIOWrapper): CSV file
        data (ndarray[<custom dtype>]): Custom dtype ndarray with column names
        headers (list[str]): Column names for custom ndarray
        row_fmt (str): Format for each row in csv
    """
    if len(headers) != len(row_fmt.split(",")):
        raise RuntimeError(f"ERROR: write_ndarray_to_fmt_csv requires row_fmt ({row_fmt}) and headers ({headers}) to have the same length.")

    # Copy data as np.ravel does not work with custom dtype arrays
    # Default type of np.empty is np.float64.
    data_cpy = np.empty((data.shape[0], len(headers)))
    for i in range(len(headers)):
        data_cpy[:, i] = data[headers[i]]

    # Create one large formatted string
    final_fmt = "\n".join([row_fmt] * data_cpy.shape[0])
    str_data = final_fmt % tuple(np.ravel(data_cpy))

    output_file.write(str_data)
    output_file.write("\n")


float_equal_precision = np.finfo(oasis_float).eps


@nb.njit(cache=True)
def almost_equal(a, b):
    return abs(a - b) < float_equal_precision


def resolve_file(path, mode, stack):
    """Resolve file path to open file or use sys.stdin

    Args:
        path (str | os.PathLike): File path or "-" indicationg standard input/output.
        mode (str): Mode to open file ("r", "rb, "w", "wb").
        stack (ExitStack): Context manager stack used to manage file lifecycle.

    Returns:
        file (IO): A file-like object opened in the specified mode.
    """
    is_read = "r" in mode
    is_binary = "b" in mode

    if str(path) == "-":
        if is_read:
            return sys.stdin.buffer if is_binary else sys.stdin
        else:
            return sys.stdout.buffer if is_binary else sys.stdout
    else:
        return stack.enter_context(open(path, mode))



__all__ = [
    "OasisBaseCommand",
    "OasisComputationCommand",
]

import logging
import os
import sys

from argparsetree import BaseCommand
from ods_tools.oed.settings import Settings, ROOT_USER_ROLE

from ..utils.path import PathCleaner
from ..utils.inputs import InputValues

from ..manager import OasisManager as om

from ..utils.log_config import OasisLogConfig
import warnings
from typing import Dict, Any
import json


class OasisBaseCommand(BaseCommand):
    """
    The base command to inherit from for each command.

    2 additional arguments (``--verbose`` and ``--config``) are added to
    the parser so that they are available for all commands.
    """

    def __init__(self, *args, **kwargs):
        self._logger = None
        self.args = None
        self.log_verbose = False
        super(OasisBaseCommand, self).__init__(*args, **kwargs)

    def add_args(self, parser):
        """
        Adds arguments to the argument parser. This is used to modify
        which arguments are processed by the command.

        Enhanced logging arguments (--log-level, --log-format) added.
        Legacy --verbose flag maintained for backward compatibility.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        # Create temporary log config instance for dynamic choices
        log_config = OasisLogConfig()

        # Legacy verbose flag (backward compatibility with deprecation notice)
        parser.add_argument(
            "-V",
            "--verbose",
            action="store_true",
            help="Use verbose logging. (Deprecated: use --log-level=DEBUG)",
        )

        # Enhanced logging arguments
        parser.add_argument(
            "-L",
            "--log-level",
            choices=log_config.get_available_levels(),
            help="Set logging level (default: INFO)",
        )

        parser.add_argument(
            "--log-format",
            choices=log_config.get_available_formats(),
            help="Set log format template (default: standard)",
        )

        # Configuration file argument
        parser.add_argument(
            "-C",
            "--config",
            required=False,
            type=PathCleaner("MDK config. JSON file", preexists=True),
            help="MDK config. JSON file",
            default="./oasislmf.json" if os.path.isfile("./oasislmf.json") else None,
        )

    def parse_args(self):
        """
        Parses the command line arguments and sets them in ``self.args``

        :return: The arguments taken from the command line
        """
        try:
            self.args = super(OasisBaseCommand, self).parse_args()

            # Handle backward compatibility with deprecation warning
            if self.args.verbose:
                warnings.warn(
                    "The --verbose flag is deprecated and will be removed in a future version. "
                    "Use --log-level=DEBUG instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            self.setup_logger()
            return self.args
        except Exception:
            self.setup_logger()
            raise

    def _load_config_dict(self) -> Dict[str, Any]:
        """
        Load configuration dictionary from file if available.

        Returns:
            Configuration dictionary or empty dict if loading fails
        """
        if not (
            hasattr(self, "args")
            and self.args
            and hasattr(self.args, "config")
            and self.args.config
        ):
            return {}

        try:
            with open(self.args.config, "r") as f:
                config = json.load(f)
            return config
        except FileNotFoundError:
            print(
                f"Warning: Config file not found: {self.args.config}", file=sys.stderr
            )
            return {}
        except json.JSONDecodeError as e:
            print(
                f"Warning: Invalid JSON in config file {self.args.config}: {e}",
                file=sys.stderr,
            )
            return {}
        except Exception as e:
            print(
                f"Warning: Could not load config file {self.args.config}: {e}",
                file=sys.stderr,
            )
            return {}

    def setup_logger(self):
        """
        Setup logger using OasisLogConfig for enhanced logging configuration.

        Supports configurable log levels, formats, and maintains backward compatibility.
        """
        if not self._logger:
            # Load configuration and create log config handler
            config_dict = self._load_config_dict()
            log_config = OasisLogConfig(config_dict)

            # Validate configuration and show warnings
            warnings_list = log_config.validate_config()
            for warning in warnings_list:
                print(f"Warning: {warning}", file=sys.stderr)

            # Get effective log level and formatter
            cli_level = (
                getattr(self.args, "log_level", None)
                if hasattr(self, "args") and self.args
                else None
            )
            is_verbose = (
                getattr(self.args, "verbose", False)
                if hasattr(self, "args") and self.args
                else False
            )
            cli_format = (
                getattr(self.args, "log_format", None)
                if hasattr(self, "args") and self.args
                else None
            )

            log_level = log_config.get_log_level(cli_level, is_verbose)
            formatter = log_config.create_formatter(cli_format)
            ods_level = log_config.get_ods_tools_level(log_level)

            # Setup main oasislmf logger
            logger = logging.getLogger("oasislmf")

            # Remove existing handlers (preserve existing logic)
            for handler in list(logger.handlers):
                if handler.name == "oasislmf":
                    logger.removeHandler(handler)
                    break

            # Setup ods_tools logger
            ods_logger = logging.getLogger("ods_tools")
            ods_logger.setLevel(ods_level)
            ods_logger.propagate = False

            # Create and configure handler
            ch = logging.StreamHandler(stream=sys.stdout)
            ch.name = "oasislmf"
            ch.setFormatter(formatter)

            # Add handler to both loggers
            logger.addHandler(ch)
            ods_logger.addHandler(ch)
            logger.setLevel(log_level)

            # Set the logger and preserve backward compatibility
            self._logger = logger
            self.log_verbose = log_level <= logging.DEBUG

            # Add debug info when running in debug mode
            if log_level <= logging.DEBUG:
                config_source = (
                    getattr(self.args, "config", None)
                    if hasattr(self, "args") and self.args
                    else None
                )
                self._logger.debug(
                    f"Effective log level: {logging.getLevelName(log_level)}"
                )
                self._logger.debug(
                    f"ods_tools level: {logging.getLevelName(ods_level)}"
                )
                self._logger.debug(
                    f"Config source: {config_source if config_source else 'default'}"
                )

    @property
    def logger(self):
        if self._logger:
            return self._logger

        return logging.getLogger("oasislmf")


class OasisComputationCommand(OasisBaseCommand):
    """
    Eventually, the Parent class for all Oasis Computation Command
    create the command line interface from parameter define in the associated computation step
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super().add_args(parser)

        for param in om.computations_params[self.computation_name]:
            add_argument_kwargs = {
                key: param.get(key)
                for key in [
                    "action",
                    "nargs",
                    "const",
                    "type",
                    "choices",
                    "help",
                    "metavar",
                    "dest",
                ]
                if param.get(key) is not None
            }
            # If 'Help' is not set then this is a function only paramter, skip
            if "help" in add_argument_kwargs:
                arg_name = f"--{param['name'].replace('_', '-')}"
                if param.get("flag"):
                    parser.add_argument(
                        param.get("flag"), arg_name, **add_argument_kwargs
                    )
                else:
                    parser.add_argument(arg_name, **add_argument_kwargs)

    @classmethod
    def get_arguments(cls, args, manager_method):
        inputs = InputValues(args)

        def get_kwargs_item(param):
            return param["name"], inputs.get(
                param["name"],
                required=param.get("required"),
                is_path=param.get("is_path"),
                dtype=param.get("type"),
            )

        settings_args = {
            param["name"] for param in manager_method.get_params(param_type="settings")
        }

        _kwargs = dict(
            get_kwargs_item(param)
            for param in manager_method.get_params()
            if param["name"] in settings_args
        )

        # read and merge computation settings files
        computation_settings = Settings()
        computation_settings.add_settings(inputs.config, ROOT_USER_ROLE)
        for settings_info in manager_method.get_params(param_type="settings"):
            setting_fp = _kwargs.get(settings_info["name"])
            if setting_fp:
                new_settings = settings_info["loader"](setting_fp)
                computation_settings.add_settings(
                    new_settings.pop("computation_settings", {}),
                    settings_info.get("user_role"),
                )
        inputs.config = computation_settings.get_settings()

        return {
            **dict(get_kwargs_item(param) for param in manager_method.get_params()),
            **_kwargs,
        }

    def action(self, args):
        """
        Generic method that call the correct manager method from the child class computation_name

        :param args: The arguments from the command line
        :type args: Namespace
        """
        manager_method = getattr(
            om(), om.computation_name_to_method(self.computation_name)
        )
        _kwargs = self.get_arguments(args, manager_method)

        self.logger.info(f"\nStarting oasislmf command - {self.computation_name}")
        manager_method(**_kwargs)



import numpy as np
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO


def returnperiods_tobin(stack, file_in, file_out, file_type):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)
    data = np.sort(data, order="return_period")[::-1]
    data.tofile(file_out)



__all__ = [
    'MarkdownGenerator'
]


from collections import defaultdict
import re


class MarkdownGenerator:
    def __init__(self):
        """A simple markdown generator for adding markdown strings
        """
        self.sections = []

    def get_markdown(self, generate_toc=False):
        """Returns markdown string from joined self.sections
        Args:
            generate_toc (bool): Generate table of contents bool.
        Returns:
            str: Markdown string
        """
        if generate_toc:
            self.generate_toc()
        return "".join(self.sections)

    def _slugify(self, title):
        """Make title strings slugified (transform to URL friendly string)
        Args:
            title (str): Original Title str
        Returns:
            slug_title (str): Slugified Title str
        """
        slug = re.sub(r'[^\w\s-]', '', title).strip().lower()
        slug = re.sub(r'\s+', '-', slug)
        return slug

    def generate_toc(self, ):
        """Generate a table of contents from markdown string
        Returns:
            toc (str): Table of contents markdown string 
        """
        markdown_text = "".join(self.sections)
        lines = markdown_text.split('\n')
        toc = []
        slug_counts = defaultdict(int)

        for line in lines:
            match = re.match(r'^(#{1,6})\s+(.*)', line)
            if match:
                level = len(match.group(1)) - 1
                title = match.group(2).strip()
                base_slug = self._slugify(title)
                slug_counts[base_slug] += 1
                anchor = base_slug if slug_counts[base_slug] == 1 else f"{base_slug}-{slug_counts[base_slug] - 1}"
                toc.append(f"{'  ' * level}- [{title}](#{anchor})")

        self.sections = ["## Table of Contents\n\n" + "\n".join(toc) + "\n\n"] + self.sections

    def add_header(self, title, level=1):
        """Adds header to markdown
        Args:
            title (Any): Title string
            level (int): Markdown header level. Defaults to 1.
        """
        self.sections.append(f"{'#' * level} {title}\n")

    def add_definition(self, title, content):
        """Adds definition line to markdown in the following format
        **title**: content
        Args:
            title (Any): Name
            content (Any): Description
        """
        self.sections.append(f"**{title}**: {content}\n\n")

    def add_table(self, headers, rows):
        """Adds a table to markdown with headers and rows
        Args:
            headers (List[str]): Headers
            rows (List[str]): Rows
        """
        if len(rows) > 0:
            assert len(rows[0]) == len(headers), \
                f"Length of rows ({len(rows[0])}) \
                does not equal length of headers \
                ({len(headers)}) for headers:\n {headers}\n"
        table = "| " + " | ".join(headers) + " |\n"
        table += "|" + "|".join(["---"] * len(headers)) + "|\n"
        for row in rows:
            table += "| " + " | ".join(row) + " |\n"
        self.sections.append(table)
        self.sections.append("\n")

    def add_list(self, items):
        """Adds list to markdown
        Args:
            items (List[str]): List of items
        """
        for item in items:
            self.sections.append(f"- {item}\n")
        self.sections.append("\n")

    def add_collapsible_section(self, text, title="Root"):
        """Adds collapsible section to markdown
        Args:
            text (str): contents of collapsible section
            title (str, optional): Collapsible section title text. Defaults to "Root".
        """
        self.add_text(f"<details><summary>{title}</summary>\n\n```json\n" + text + "\n```\n</details>")

    def add_text(self, content):
        """Adds text to markdown
        Args:
            content (Any): Text content
        """
        self.sections.append(f"{content}\n\n")



#!/usr/bin/env python

from contextlib import ExitStack
import logging

from oasislmf.pytools.common.data import resolve_file
from oasislmf.pytools.converters.csvtobin.utils import (
    amplifications_tobin,
    complex_items_tobin,
    coverages_tobin,
    damagebin_tobin,
    fm_tobin,
    footprint_tobin,
    gul_tobin,
    lossfactors_tobin,
    occurrence_tobin,
    returnperiods_tobin,
    summarycalc_tobin,
    vulnerability_tobin,
)
from oasislmf.pytools.converters.csvtobin.utils.common import read_csv_as_ndarray
from oasislmf.pytools.converters.data import TOOL_INFO

logger = logging.getLogger(__name__)


TOBIN_FUNC_MAP = {
    "amplifications": amplifications_tobin,
    "complex_items": complex_items_tobin,
    "coverages": coverages_tobin,
    "damagebin": damagebin_tobin,
    "fm": fm_tobin,
    "footprint": footprint_tobin,
    "gul": gul_tobin,
    "lossfactors": lossfactors_tobin,
    "occurrence": occurrence_tobin,
    "returnperiods": returnperiods_tobin,
    "summarycalc": summarycalc_tobin,
    "vulnerability": vulnerability_tobin,
}


def default_tobin(stack, file_in, file_out, file_type):
    headers = TOOL_INFO[file_type]["headers"]
    dtype = TOOL_INFO[file_type]["dtype"]
    data = read_csv_as_ndarray(stack, file_in, headers, dtype)
    data.tofile(file_out)


def csvtobin(file_in, file_out, file_type, **kwargs):
    """Convert csv file to bin file based on file type
    Args:
        file_in (str | os.PathLike): Input file path
        file_out (str | os.PathLike): Output file path
        file_type (str): File type str from SUPPORTED_CSVTOBIN
    """
    with ExitStack() as stack:
        file_out = resolve_file(file_out, "wb", stack)

        tobin_func = TOBIN_FUNC_MAP.get(file_type, default_tobin)
        tobin_func(stack, file_in, file_out, file_type, **kwargs)



import importlib
import pkgutil


TOOL_MODULE_PACKAGE = "oasislmf.pytools.converters.tools"


def get_tools_by_cli(tools_info: dict, cli_command: str) -> list[str]:
    """Return a list of tool names supported by the given CLI command."""
    return [tool for tool, info in tools_info.items() if cli_command in info.get("cli_support", [])]


def build_tool_info():
    """ Generates the tools information dictionary for conversion tools from the files in TOOL_MODULE_PACKAGE

    Returns:
        tool_indo (dict): TOOL_INFO dict
    """
    tool_info = {}

    package = importlib.import_module(TOOL_MODULE_PACKAGE)
    for _, module_name, _ in pkgutil.iter_modules(package.__path__):
        mod = importlib.import_module(f"{TOOL_MODULE_PACKAGE}.{module_name}")

        tool_info[module_name] = {
            "headers": getattr(mod, "headers", None),
            "dtype": getattr(mod, "dtype", None),
            "fmt": getattr(mod, "fmt", None),
            "cli_support": getattr(mod, "cli_support", []),
        }

    return tool_info


TOOL_INFO = build_tool_info()
SUPPORTED_BINTOCSV = get_tools_by_cli(TOOL_INFO, "bintocsv")
SUPPORTED_CSVTOBIN = get_tools_by_cli(TOOL_INFO, "csvtobin")
SUPPORTED_BINTOPARQUET = get_tools_by_cli(TOOL_INFO, "bintoparquet")
SUPPORTED_PARQUETTOBIN = get_tools_by_cli(TOOL_INFO, "parquettobin")



import numpy as np

from oasislmf.pytools.common.data import oasis_int, oasis_float


VALID_EXT = ["csv", "bin", "parquet"]

SPLT_output = [
    ('Period', oasis_int, '%d'),
    ('PeriodWeight', oasis_float, '%.6f'),
    ('EventId', oasis_int, '%d'),
    ('Year', oasis_int, '%d'),
    ('Month', oasis_int, '%d'),
    ('Day', oasis_int, '%d'),
    ('Hour', oasis_int, '%d'),
    ('Minute', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('SampleId', oasis_int, '%d'),
    ('Loss', oasis_float, '%.2f'),
    ('ImpactedExposure', oasis_float, '%.2f'),
]

MPLT_output = [
    ('Period', oasis_int, '%d'),
    ('PeriodWeight', oasis_float, '%.6f'),
    ('EventId', oasis_int, '%d'),
    ('Year', oasis_int, '%d'),
    ('Month', oasis_int, '%d'),
    ('Day', oasis_int, '%d'),
    ('Hour', oasis_int, '%d'),
    ('Minute', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('SampleType', oasis_int, '%d'),
    ('ChanceOfLoss', oasis_float, '%.4f'),
    ('MeanLoss', oasis_float, '%.2f'),
    ('SDLoss', oasis_float, '%.2f'),
    ('MaxLoss', oasis_float, '%.2f'),
    ('FootprintExposure', oasis_float, '%.2f'),
    ('MeanImpactedExposure', oasis_float, '%.2f'),
    ('MaxImpactedExposure', oasis_float, '%.2f'),
]

QPLT_output = [
    ('Period', oasis_int, '%d'),
    ('PeriodWeight', oasis_float, '%.6f'),
    ('EventId', oasis_int, '%d'),
    ('Year', oasis_int, '%d'),
    ('Month', oasis_int, '%d'),
    ('Day', oasis_int, '%d'),
    ('Hour', oasis_int, '%d'),
    ('Minute', oasis_int, '%d'),
    ('SummaryId', oasis_int, '%d'),
    ('Quantile', oasis_float, '%.2f'),
    ('Loss', oasis_float, '%.2f'),
]

SPLT_headers = [c[0] for c in SPLT_output]
MPLT_headers = [c[0] for c in MPLT_output]
QPLT_headers = [c[0] for c in QPLT_output]
SPLT_dtype = np.dtype([(c[0], c[1]) for c in SPLT_output])
MPLT_dtype = np.dtype([(c[0], c[1]) for c in MPLT_output])
QPLT_dtype = np.dtype([(c[0], c[1]) for c in QPLT_output])
SPLT_fmt = ','.join([c[2] for c in SPLT_output])
MPLT_fmt = ','.join([c[2] for c in MPLT_output])
QPLT_fmt = ','.join([c[2] for c in QPLT_output])



# katpy/manager.py

from collections import Counter
from contextlib import ExitStack
import csv
import glob
import heapq
import logging
import numba as nb
import numpy as np
import shutil
from pathlib import Path
import pyarrow as pa
import pyarrow.parquet as pq
import tempfile

from oasislmf.pytools.common.data import DEFAULT_BUFFER_SIZE, write_ndarray_to_fmt_csv
from oasislmf.pytools.common.utils.nb_heapq import heap_pop, heap_push, init_heap
from oasislmf.pytools.elt.data import MELT_dtype, MELT_fmt, MELT_headers, QELT_dtype, QELT_fmt, QELT_headers, SELT_dtype, SELT_fmt, SELT_headers
from oasislmf.pytools.plt.data import MPLT_dtype, MPLT_fmt, MPLT_headers, QPLT_dtype, QPLT_fmt, QPLT_headers, SPLT_dtype, SPLT_fmt, SPLT_headers
from oasislmf.pytools.utils import redirect_logging

logger = logging.getLogger(__name__)


KAT_SELT = 0
KAT_MELT = 1
KAT_QELT = 2
KAT_SPLT = 3
KAT_MPLT = 4
KAT_QPLT = 5

KAT_MAP = {
    KAT_SELT: {
        "name": "SELT",
        "headers": SELT_headers,
        "dtype": SELT_dtype,
        "fmt": SELT_fmt,
    },
    KAT_MELT: {
        "name": "MELT",
        "headers": MELT_headers,
        "dtype": MELT_dtype,
        "fmt": MELT_fmt,
    },
    KAT_QELT: {
        "name": "QELT",
        "headers": QELT_headers,
        "dtype": QELT_dtype,
        "fmt": QELT_fmt,
    },
    KAT_SPLT: {
        "name": "SPLT",
        "headers": SPLT_headers,
        "dtype": SPLT_dtype,
        "fmt": SPLT_fmt,
    },
    KAT_MPLT: {
        "name": "MPLT",
        "headers": MPLT_headers,
        "dtype": MPLT_dtype,
        "fmt": MPLT_fmt,
    },
    KAT_QPLT: {
        "name": "QPLT",
        "headers": QPLT_headers,
        "dtype": QPLT_dtype,
        "fmt": QPLT_fmt,
    },
}


def check_file_extensions(file_paths):
    """Check file path extensions are all identical
    Args:
        file_paths (List[str | os.PathLike]): List of csv file paths.
    Returns:
        ext (str): file extension as a str
    """
    first_ext = file_paths[0].suffix

    if all(fp.suffix == first_ext for fp in file_paths):
        return first_ext
    raise RuntimeError("ERROR: katpy has input files with different file extensions. Make sure all input files are of the same type.")


def find_csv_with_header(
    stack,
    file_paths,
):
    """Find and check csv files for consistent and present headers
    Args:
        stack (ExitStack): Exit Stack.
        file_paths (List[str | os.PathLike]): List of csv file paths.
    Returns:
        files_with_header (List[bool]): Bool list of files with header present
        header (str): Header to write
    """
    headers = {}
    files_with_header = [False] * len(file_paths)
    for idx, fp in enumerate(file_paths):
        with stack.enter_context(fp.open("r", newline="", encoding="utf-8")) as f:
            first_row = f.readline().strip()
            if first_row:
                if "EventId" in first_row:
                    headers[fp] = first_row
                    files_with_header[idx] = True

    if not headers:
        raise ValueError("ERROR: katpy, no valid header found in any CSV file.")

    header_counts = Counter(headers.values())

    if len(header_counts) > 1:
        raise ValueError(f"ERROR: katpy, conflicting headers found: {header_counts}")

    return files_with_header, list(headers.values())[0]


def check_correct_headers(headers, file_type):
    """Checks headers found in csv file matches excpected headers for file type
    Args:
        headers (List[str]): Headers
        file_type (int): File type int matching KAT_NAMES index
    """
    if file_type in KAT_MAP:
        expected_headers = KAT_MAP[file_type]["headers"]
    else:
        file_type_names = [v["name"] for v in KAT_MAP.values()]
        raise ValueError(f"ERROR: katpy, unknown file_type {file_type}, not in {file_type_names}")
    if headers != expected_headers:
        raise RuntimeError(f"ERROR: katpy, incorrect headers found in csv file, expected {expected_headers} but got {headers}")


def get_header_idxs(
    headers,
    headers_to_search,
):
    """Search for index of headers_to_search in headers of csv file
    Args:
        headers (List[str]): Headers
        headers_to_search (List[str]): Headers to search
    Returns:
        idxs (List[int]): Indexes of searched headers
    """
    idxs = []
    for header in headers_to_search:
        try:
            idxs.append(headers.index(header))
        except ValueError as e:
            raise ValueError(f"ERROR: katpy, cannot sort by header {header}, not found. {e}")
    return idxs


def csv_concat_unsorted(
    stack,
    file_paths,
    files_with_header,
    headers,
    out_file,
):
    """Concats CSV files in order they are passed in.
    Args:
        stack (ExitStack): Exit Stack.
        file_paths (List[str | os.PathLike]): List of csv file paths.
        files_with_header (List[bool]): Bool list of files with header present
        headers (List[str]): Headers to write
        out_file (str | os.PathLike): Output Concatenated CSV file.
    """
    first_header_written = False

    with stack.enter_context(out_file.open("wb")) as out:
        for i, fp in enumerate(file_paths):
            with stack.enter_context(fp.open("rb")) as csv_file:
                if files_with_header[i]:
                    # Read first line (header)
                    first_line = csv_file.readline()

                    # Write the expected header at the start of the file
                    if not first_header_written:
                        out.write(",".join(headers).encode() + b"\n")
                        first_header_written = True
                shutil.copyfileobj(csv_file, out)


def csv_concat_sort_by_headers(
    stack,
    file_paths,
    files_with_header,
    headers,
    header_idxs,
    sort_fn,
    out_file,
    **sort_kwargs
):
    """Concats CSV files in order determined by the header_idxs and sort_fn
    Args:
        stack (ExitStack): Exit Stack.
        file_paths (List[str | os.PathLike]): List of csv file paths.
        files_with_header (List[bool]): Bool list of files with header present
        headers (List[str]): Headers to write
        header_idxs (List[int]): Indices of headers to sort by
        sort_fn (Callable[[List[int]], Any]): Sort function to apply to header_idxs
        out_file (str | os.PathLike): Output Concatenated CSV file.
    """
    # Open all csv files
    csv_files = [stack.enter_context(open(fp, "r", newline="")) for fp in file_paths]
    readers = [csv.reader(f) for f in csv_files]

    # Skip headers
    for idx, reader in enumerate(readers):
        if files_with_header[idx]:
            next(reader, None)

    def row_generator(reader, file_idx):
        for row in reader:
            sort_key = sort_fn([row[i] for i in header_idxs], **sort_kwargs)
            yield (sort_key, file_idx, row)

    # Create iterator for each file
    iterators = [row_generator(reader, idx) for idx, reader in enumerate(readers)]

    # Merge all iterators
    merged_iterator = heapq.merge(*iterators, key=lambda x: x[0])

    buffer_size = DEFAULT_BUFFER_SIZE
    with stack.enter_context(out_file.open("w")) as out:
        writer = csv.writer(out, lineterminator="\n")
        writer.writerow(headers)

        buffer = []
        for _, _, row in merged_iterator:
            buffer.append(row)
            if len(buffer) > buffer_size:
                writer.writerows(buffer)
                buffer.clear()

        if buffer:
            writer.writerows(buffer)

    for f in csv_files:
        f.close()


def bin_concat_unsorted(
    stack,
    file_paths,
    out_file,
):
    """Concats Binary files in order they are passed in.
    Args:
        stack (ExitStack): Exit Stack.
        file_paths (List[str | os.PathLike]): List of bin file paths.
        out_file (str | os.PathLike): Output Concatenated Binary file.
    """
    with stack.enter_context(out_file.open('wb')) as out:
        for fp in file_paths:
            with fp.open('rb') as bin_file:
                shutil.copyfileobj(bin_file, out)


@nb.njit(cache=True, error_model="numpy")
def merge_elt_data(memmaps):
    """Merge sorted chunks using a k-way merge algorithm
    Args:
        memmaps (List[np.memmap]): List of temporary file memmaps
    Yields:
        buffer (ndarray): yields sorted buffer from memmaps
    """
    min_heap = init_heap(num_compare=1)
    size = 0
    # Initialize the min_heap with the first row of each memmap
    for i, mmap in enumerate(memmaps):
        if len(mmap) > 0:
            first_row = mmap[0]
            min_heap, size = heap_push(min_heap, size, np.array(
                [first_row["EventId"], i, 0], dtype=np.int32
            ))

    buffer_size = DEFAULT_BUFFER_SIZE
    buffer = np.empty(buffer_size, dtype=memmaps[0].dtype)
    bidx = 0

    # Perform the k-way merge
    while size > 0:
        # The min heap will store the smallest row at the top when popped
        element, min_heap, size = heap_pop(min_heap, size)
        file_idx = element[-2]
        row_num = element[-1]
        smallest_row = memmaps[file_idx][row_num]

        # Add to buffer and yield when full
        buffer[bidx] = smallest_row
        bidx += 1
        if bidx >= buffer_size:
            yield buffer[:bidx]
            bidx = 0

        # Push the next row from the same file into the heap if there are any more rows
        if row_num + 1 < len(memmaps[file_idx]):
            next_row = memmaps[file_idx][row_num + 1]
            min_heap, size = heap_push(min_heap, size, np.array(
                [next_row["EventId"], file_idx, row_num + 1], dtype=np.int32
            ))
    yield buffer[:bidx]


@nb.njit(cache=True, error_model="numpy")
def merge_plt_data(memmaps):
    """Merge sorted chunks using a k-way merge algorithm
    Args:
        memmaps (List[np.memmap]): List of temporary file memmaps
    Yields:
        buffer (ndarray): yields sorted buffer from memmaps
    """
    min_heap = init_heap(num_compare=2)
    size = 0
    # Initialize the min_heap with the first row of each memmap
    for i, mmap in enumerate(memmaps):
        if len(mmap) > 0:
            first_row = mmap[0]
            min_heap, size = heap_push(min_heap, size, np.array(
                [first_row["EventId"], first_row["Period"], i, 0], dtype=np.int32
            ))

    buffer_size = DEFAULT_BUFFER_SIZE
    buffer = np.empty(buffer_size, dtype=memmaps[0].dtype)
    bidx = 0

    # Perform the k-way merge
    while size > 0:
        # The min heap will store the smallest row at the top when popped
        element, min_heap, size = heap_pop(min_heap, size)
        file_idx = element[-2]
        row_num = element[-1]
        smallest_row = memmaps[file_idx][row_num]

        # Add to buffer and yield when full
        buffer[bidx] = smallest_row
        bidx += 1
        if bidx >= buffer_size:
            yield buffer[:bidx]
            bidx = 0

        # Push the next row from the same file into the heap if there are any more rows
        if row_num + 1 < len(memmaps[file_idx]):
            next_row = memmaps[file_idx][row_num + 1]
            min_heap, size = heap_push(min_heap, size, np.array(
                [next_row["EventId"], next_row["Period"], file_idx, row_num + 1], dtype=np.int32
            ))
    yield buffer[:bidx]


def bin_concat_sort_by_headers(
    stack,
    file_paths,
    file_type,
    out_type,
    out_file,
):
    """Concats Binary files in order determined out_type and their respective merge functions
    Args:
        stack (ExitStack): Exit Stack.
        file_paths (List[str | os.PathLike]): List of bin file paths.
        file_type (int): File type int matching KAT_NAMES index
        out_type (str): Out type str between "elt" and "plt"
        out_file (str | os.PathLike): Output Concatenated Binary file.
    """
    files = [np.memmap(fp, dtype=KAT_MAP[file_type]["dtype"]) for fp in file_paths]

    if out_type == "elt":
        gen = merge_elt_data(files)
    elif out_type == "plt":
        gen = merge_plt_data(files)
    else:
        raise RuntimeError(f"ERROR: katpy, unknown out_type {out_type}")

    with stack.enter_context(out_file.open("wb")) as out:
        for data in gen:
            data.tofile(out)


def parquet_concat_unsorted(
    file_paths,
    out_file,
):
    """Concats Parquet files in order they are passed in.
    Args:
        file_paths (List[str | os.PathLike]): List of parquet file paths.
        out_file (str | os.PathLike): Output Concatenated Parquet file.
    """
    writer = None
    for fp in file_paths:
        pq_file = pq.ParquetFile(fp)
        for rg in range(pq_file.num_row_groups):
            table = pq_file.read_row_group(rg)
            if writer is None:
                writer = pq.ParquetWriter(out_file, table.schema)
            writer.write_table(table)

    if writer:
        writer.close()


def parquet_kway_merge(
    file_paths,
    keys,
    chunk_size=100000,
):
    """Merge sorted chunks using a k-way merge algorithm
    Args:
        file_paths (List[str | os.PathLike]): List of parquet file paths.
        keys (List[str]): List of keys to sort by
        chunk_size (int): Chunk size for reading parquet files. Defaults to 100000.
    Yields:
        buffer (pa.Table): yields sorted pyarrow table from input files
    """
    # Helper to read batches and manage current state for each file
    class FileStream:
        def __init__(self, path):
            self.reader = pq.ParquetFile(str(path)).iter_batches(batch_size=chunk_size)
            self.chunk = None
            self.table = None
            self.index = 0
            self._load_next_chunk()

        def _load_next_chunk(self):
            try:
                self.chunk = next(self.reader)
                self.table = self.chunk.to_pydict()
                self.index = 0
            except StopIteration:
                self.chunk = None
                self.table = None

        def has_data(self):
            return self.table is not None

        def current_key(self, keys):
            return tuple(self.table[k][self.index] for k in keys)

        def current_row(self):
            return {k: self.table[k][self.index] for k in self.table}

        def advance(self):
            self.index += 1
            if self.index >= len(next(iter(self.table.values()))):
                self._load_next_chunk()

    streams = [FileStream(fp) for fp in file_paths]
    heap = [
        (stream.current_key(keys), i)
        for i, stream in enumerate(streams) if stream.has_data()
    ]
    heapq.heapify(heap)

    buffer = {}
    schema = None

    while heap:
        _, i = heapq.heappop(heap)
        stream = streams[i]

        # Initialize schema and buffer
        if schema is None:
            schema = stream.chunk.schema
            buffer = {name: [] for name in schema.names}

        # Append current row to buffer
        row = stream.current_row()
        for col in buffer:
            buffer[col].append(row[col])

        stream.advance()
        if stream.has_data():
            heapq.heappush(heap, (stream.current_key(keys), i))

        # Output buffer if full
        if len(buffer[keys[0]]) >= chunk_size:
            yield pa.table(buffer, schema=schema)
            buffer = {name: [] for name in schema.names}

    # Yield remaining buffer
    if buffer and buffer[keys[0]]:
        yield pa.table(buffer, schema=schema)


def parquet_concat_sorted(
    file_paths,
    out_type,
    out_file,
    chunk_size=100000,
):
    """Concats Parquet files in order determined out_type and their respective merge functions
    Args:
        file_paths (List[str | os.PathLike]): List of parquet file paths.
        out_type (str): Out type str between "elt" and "plt"
        out_file (str | os.PathLike): Output Concatenated Parquet file.
        chunk_size (int): Chunk size for reading parquet files. Defaults to 100000.
    """
    if out_type == "elt":
        keys = ["EventId"]
    elif out_type == "plt":
        keys = ["EventId", "Period"]
    else:
        raise RuntimeError(f"Unknown out_type: {out_type}")

    writer = None

    for table in parquet_kway_merge(file_paths, keys, chunk_size):
        if writer is None:
            writer = pq.ParquetWriter(out_file, table.schema)
        writer.write_table(table)

    if writer:
        writer.close()


def run(
    out_file,
    file_type=None,
    files_in=None,
    dir_in=None,
    concatenate_selt=False,
    concatenate_melt=False,
    concatenate_qelt=False,
    concatenate_splt=False,
    concatenate_mplt=False,
    concatenate_qplt=False,
    unsorted=False,
):
    """Concatenate CSV files (optionally sorted)
    Args:
        out_file (str | os.PathLike): Output Concatenated CSV file.
        file_type (str, optional): Input file type suffix, if not discernible from input files. Defaults to None.
        files_in (List[str | os.PathLike], optional): Individual CSV file paths to concatenate. Defaults to None.
        dir_in (str | os.PathLike, optional): Path to the directory containing files for concatenation. Defaults to None.
        concatenate_selt (bool, optional): Concatenate SELT CSV file. Defaults to False.
        concatenate_melt (bool, optional): Concatenate MELT CSV file. Defaults to False.
        concatenate_qelt (bool, optional): Concatenate QELT CSV file. Defaults to False.
        concatenate_splt (bool, optional): Concatenate SPLT CSV file. Defaults to False.
        concatenate_mplt (bool, optional): Concatenate MPLT CSV file. Defaults to False.
        concatenate_qplt (bool, optional): Concatenate QPLT CSV file. Defaults to False.
        unsorted (bool, optional): Do not sort by event/period ID. Defaults to False.
    """
    input_files = []

    # Check and add files from dir_in
    if dir_in:
        dir_in = Path(dir_in)
        if not dir_in.exists():
            raise FileNotFoundError(f"ERROR: Directory \'{dir_in}\' does not exist")
        if not dir_in.is_dir():
            raise ValueError(f"ERROR: \'{dir_in}\' is not a directory.")

        dir_csv_input_files = glob.glob(str(dir_in / "*.csv"))
        dir_bin_input_files = glob.glob(str(dir_in / "*.bin"))
        dir_parquet_input_files = glob.glob(str(dir_in / "*.parquet"))
        if not dir_csv_input_files and not dir_bin_input_files and not dir_parquet_input_files:
            logger.warning(f"Warning: No valid files found in directory \'{dir_in}\'")
        input_files += [Path(file).resolve() for file in dir_csv_input_files + dir_bin_input_files + dir_parquet_input_files]

    input_files.sort()

    # Check and add files from files_in
    if files_in:
        for file in files_in:
            path = Path(file).resolve()
            if not path.exists():
                raise FileNotFoundError(f"ERROR: File \'{path}\' does not exist.")
            if not path.is_file():
                raise FileNotFoundError(f"ERROR: File \'{path}\' is not a valid file.")
            input_files.append(path)

    if not input_files:
        raise RuntimeError("ERROR: katpy has no input CSV files to join")

    out_file = Path(out_file).resolve()
    input_type = check_file_extensions(input_files)
    output_type = out_file.suffix

    if input_type == "":
        if not file_type:
            raise RuntimeError("ERROR: katpy, no discernible file type suffix found from input files, please provide a file_type")
        input_type = "." + file_type

    # If out_file is a csv and input_files are not csvs, then output to temporary outfile
    # of type input_type, and convert to csv after
    bin_to_csv = False
    bin_to_parquet = False
    if output_type != input_type:
        if input_type == ".bin":
            if output_type not in [".csv", ".parquet"]:
                raise RuntimeError(f"ERROR: katpy does not support concatenating input files of type {input_type} to output type {output_type}")
            final_out_file_path = out_file
            temp_file = tempfile.NamedTemporaryFile(suffix=input_type, delete=False)
            out_file = Path(temp_file.name)
            bin_to_csv = output_type == ".csv"
            bin_to_parquet = output_type == ".parquet"
        else:
            raise RuntimeError(f"ERROR: katpy does not support concatenating input files of type {input_type} to output type {output_type}")

    output_flags = [
        concatenate_selt,
        concatenate_melt,
        concatenate_qelt,
        concatenate_splt,
        concatenate_mplt,
        concatenate_qplt,
    ]

    sort_by_event = any(output_flags[KAT_SELT:KAT_QELT + 1])
    sort_by_period = any(output_flags[KAT_SPLT:KAT_QPLT + 1])
    assert sort_by_event + sort_by_period == 1, "incorrect flag config in katpy"
    out_type = output_flags.index(True)

    with ExitStack() as stack:
        if input_type == ".csv":
            files_with_header, header = find_csv_with_header(stack, input_files)
            headers = header.strip().split(",")
            check_correct_headers(headers, out_type)

            if unsorted:
                csv_concat_unsorted(stack, input_files, files_with_header, headers, out_file)
            elif sort_by_event:
                header_idxs = get_header_idxs(headers, ["EventId"])
                csv_concat_sort_by_headers(
                    stack,
                    input_files,
                    files_with_header,
                    headers,
                    header_idxs,
                    lambda values: int(values[0]),
                    out_file,
                )
            elif sort_by_period:
                header_idxs = get_header_idxs(headers, ["EventId", "Period"])
                csv_concat_sort_by_headers(
                    stack,
                    input_files,
                    files_with_header,
                    headers,
                    header_idxs,
                    lambda values: (int(values[0]), int(values[1])),
                    out_file,
                )
        elif input_type == ".bin":
            if unsorted:
                bin_concat_unsorted(stack, input_files, out_file)
            else:
                sort_type = "elt" if sort_by_event else "plt"
                bin_concat_sort_by_headers(
                    stack,
                    input_files,
                    out_type,
                    sort_type,
                    out_file,
                )
        elif input_type == ".parquet":
            if unsorted:
                parquet_concat_unsorted(
                    input_files,
                    out_file,
                )
            else:
                sort_type = "elt" if sort_by_event else "plt"
                parquet_concat_sorted(
                    input_files,
                    sort_type,
                    out_file,
                )
        else:
            raise RuntimeError(f"ERROR: katpy, file type {input_type} not supported.")

    if bin_to_csv or bin_to_parquet:
        data = np.memmap(out_file, dtype=KAT_MAP[out_type]["dtype"])
        headers = KAT_MAP[out_type]["headers"]
        fmt = KAT_MAP[out_type]["fmt"]
        num_rows = data.shape[0]
        if bin_to_csv:
            csv_out_file = open(final_out_file_path, "w")
            csv_out_file.write(",".join(headers) + "\n")

            buffer_size = DEFAULT_BUFFER_SIZE
            for start in range(0, num_rows, buffer_size):
                end = min(start + buffer_size, num_rows)
                buffer_data = data[start:end]
                write_ndarray_to_fmt_csv(csv_out_file, buffer_data, headers, fmt)
            csv_out_file.close()
        if bin_to_parquet:
            parquet_writer = None
            buffer_size = DEFAULT_BUFFER_SIZE
            for start in range(0, num_rows, buffer_size):
                end = min(start + buffer_size, num_rows)
                buffer_data = data[start:end]

                arrays = []
                fields = []
                for name in buffer_data.dtype.names:
                    array = pa.array(buffer_data[name])
                    arrays.append(array)
                    fields.append((name, array.type))

                schema = pa.schema(fields)
                table = pa.Table.from_arrays(arrays, schema=schema)
                if parquet_writer is None:
                    parquet_writer = pq.ParquetWriter(final_out_file_path, schema)
                parquet_writer.write_table(table)

            if parquet_writer is not None:
                parquet_writer.close()


@redirect_logging(exec_name='katpy')
def main(
    out=None,
    file_type=None,
    files_in=None,
    dir_in=None,
    selt=False,
    melt=False,
    qelt=False,
    splt=False,
    mplt=False,
    qplt=False,
    unsorted=False,
    **kwargs
):
    run(
        out_file=out,
        file_type=file_type,
        files_in=files_in,
        dir_in=dir_in,
        concatenate_selt=selt,
        concatenate_melt=melt,
        concatenate_qelt=qelt,
        concatenate_splt=splt,
        concatenate_mplt=mplt,
        concatenate_qplt=qplt,
        unsorted=unsorted,
    )



from oasislmf.pytools.plt.data import MPLT_headers, MPLT_dtype, MPLT_fmt


headers = MPLT_headers
dtype = MPLT_dtype
fmt = MPLT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



import logging
import os
from typing import Optional, Dict, Any, Union, List


class OasisLogConfig:
    """
    Configuration handler for OasisLMF CLI/console logging.

    Handles log level resolution, format management, and validation for CLI-based logging.
    Designed to work alongside existing file-based logging in log.py.

    Environment Variables:
        OASISLMF_LOG_LEVEL: Override log level (e.g., 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')

    Example:
        >>> # Basic usage with config file
        >>> config = OasisLogConfig({'logging': {'level': 'INFO', 'format': 'compact'}})
        >>> formatter = config.create_formatter()
        >>> level = config.get_log_level()

        >>> # CLI override example
        >>> level = config.get_log_level('DEBUG')  # Override config file
        >>> available_formats = config.get_available_formats()
        >>> available_levels = config.get_available_levels()
    """

    # Standard levels for validation and help text
    STANDARD_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

    FORMAT_TEMPLATES = {
        "simple": "%(message)s",
        "standard": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        "detailed": "%(asctime)s - %(processName)s-%(process)d - %(name)s - %(levelname)s - %(message)s",
        "iso_timestamp": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        "production": "%(asctime)s [%(process)d] %(name)s - %(levelname)s - %(message)s",
        "compact": "%(asctime)s [%(levelname)s] %(message)s",
    }

    # Date format configurations for different templates
    DATE_FORMATS = {
        "iso_timestamp": "%Y-%m-%dT%H:%M:%S",
        "standard": "%Y-%m-%d %H:%M:%S",
        "detailed": "%Y-%m-%d %H:%M:%S",
        "production": "%Y-%m-%d %H:%M:%S",
        "compact": "%H:%M:%S",
    }

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Initialize logging configuration manager.

        Args:
            config: Configuration dictionary (typically from JSON file loaded by command.py)
        """
        self.config = config or {}

    def get_log_level(
        self, cli_level: Optional[str] = None, is_verbose: bool = False
    ) -> int:
        """
        Get effective log level from various sources.

        Priority order: CLI args > env vars > config file > verbose flag > default

        Args:
            cli_level: Log level from command line (e.g., 'INFO', 'DEBUG')
            is_verbose: Legacy verbose flag for backward compatibility

        Returns:
            Numeric log level (e.g., 20 for INFO, 10 for DEBUG)

        Examples:
            >>> config = OasisLogConfig()
            >>> config.get_log_level('DEBUG')
            10
            >>> config.get_log_level(is_verbose=True)
            10
        """
        # 1. CLI argument takes highest priority
        if cli_level:
            return self._parse_level(cli_level)

        # 2. Environment variable
        env_level = os.environ.get("OASISLMF_LOG_LEVEL")
        if env_level:
            return self._parse_level(env_level)

        # 3. Config file
        config_level = self.config.get("logging", {}).get("level")
        if config_level:
            return self._parse_level(config_level)

        # 4. Verbose flag (backward compatibility)
        if is_verbose:
            return logging.DEBUG

        # 5. Default
        return logging.INFO

    def get_ods_tools_level(self, main_level: int) -> int:
        """
        Get ods_tools logger level based on main logger level.

        Args:
            main_level: Main oasislmf logger level to use as reference

        Returns:
            Numeric log level for ods_tools logger
        """
        # Check if explicitly configured
        config_level = self.config.get("logging", {}).get("ods_tools_level")
        if config_level:
            return self._parse_level(config_level)

        # Default behavior: WARNING unless main level is DEBUG
        return logging.DEBUG if main_level <= logging.DEBUG else logging.WARNING

    def get_format_string(self, format_name: Optional[str] = None) -> str:
        """
        Get log format string.

        Args:
            format_name: Format template name from CLI or None for config/default

        Returns:
            Format string for logging.Formatter
        """
        # CLI format name takes priority
        if format_name and format_name in self.FORMAT_TEMPLATES:
            return self.FORMAT_TEMPLATES[format_name]

        # Check config file
        config_format = self.config.get("logging", {}).get("format")
        if config_format:
            if config_format in self.FORMAT_TEMPLATES:
                return self.FORMAT_TEMPLATES[config_format]
            else:
                # Custom format string from config
                return config_format

        # Default: standard format
        return self.FORMAT_TEMPLATES["standard"]

    def get_date_format(self, format_name: Optional[str] = None) -> Optional[str]:
        """
        Get date format string for the specified template.

        Args:
            format_name: Format template name

        Returns:
            Date format string or None for logging default
        """
        # Check CLI format name first
        if format_name and format_name in self.DATE_FORMATS:
            return self.DATE_FORMATS[format_name]

        # Check config format
        config_format = self.config.get("logging", {}).get("format")
        if config_format and config_format in self.DATE_FORMATS:
            return self.DATE_FORMATS[config_format]

        # Default: None (uses logging module default)
        return None

    def create_formatter(self, format_name: Optional[str] = None) -> logging.Formatter:
        """
        Create a logging formatter with appropriate format and date format.

        Args:
            format_name: Format template name from CLI

        Returns:
            Configured logging.Formatter instance
        """
        format_str = self.get_format_string(format_name)
        date_format = self.get_date_format(format_name)

        return logging.Formatter(format_str, datefmt=date_format)

    def get_available_formats(self) -> List[str]:
        """
        Get list of available format template names.

        Returns:
            List of format template names that can be used with get_format_string()

        Example:
            >>> config = OasisLogConfig()
            >>> formats = config.get_available_formats()
            >>> 'standard' in formats
            True
        """
        return list(self.FORMAT_TEMPLATES.keys())

    def get_available_levels(self) -> List[str]:
        """
        Get list of available log level names.

        Returns:
            List of standard log level names that can be used with get_log_level()

        Example:
            >>> config = OasisLogConfig()
            >>> levels = config.get_available_levels()
            >>> 'DEBUG' in levels
            True
        """
        return self.STANDARD_LEVELS.copy()

    def validate_config(self) -> List[str]:
        """
        Validate logging configuration and return any issues.

        Returns:
            List of warning messages about configuration issues
        """
        warnings = []
        logging_config = self.config.get("logging", {})

        # Validate main log level
        level = logging_config.get("level")
        if level:
            try:
                self._parse_level(level)
            except ValueError as e:
                warnings.append(f"Invalid log level in config: {e}")

        # Validate ods_tools level
        ods_level = logging_config.get("ods_tools_level")
        if ods_level:
            try:
                self._parse_level(ods_level)
            except ValueError as e:
                warnings.append(f"Invalid ods_tools log level in config: {e}")

        # Validate format (lenient - custom format strings are allowed)
        format_name = logging_config.get("format")
        if format_name and format_name not in self.FORMAT_TEMPLATES:
            # Only warn if it doesn't look like a format string
            if not ("%(message)s" in str(format_name) or "%(" in str(format_name)):
                available = ", ".join(self.get_available_formats())
                warnings.append(
                    f"Format '{format_name}' may not be a valid format template. "
                    f"Available templates: {available}"
                )

        return warnings

    def _parse_level(self, level: Union[str, int, None]) -> int:
        """
        Parse log level using Python's logging module directly.

        Args:
            level: Log level as string ('DEBUG'), integer (10), or None

        Returns:
            Numeric log level

        Raises:
            ValueError: If level is invalid
        """
        if level is None:
            return logging.INFO

        if isinstance(level, str):
            # Use logging module's own level names
            numeric_level = getattr(logging, level.upper(), None)
            if numeric_level is None:
                available = ", ".join(self.get_available_levels())
                raise ValueError(
                    f"Invalid log level: '{level}'. " f"Available levels: {available}"
                )
            return numeric_level

        elif isinstance(level, int):
            # Validate integer is in reasonable range
            if 0 <= level <= 50:  # Standard logging range
                return level
            else:
                raise ValueError(
                    f"Invalid log level: {level}. Use 0-50 or standard level names."
                )

        else:
            raise ValueError(f"Log level must be string or integer, got {type(level)}")



from oasislmf.pytools.lec.data import PSEPT_headers, PSEPT_dtype, PSEPT_fmt


headers = PSEPT_headers
dtype = PSEPT_dtype
fmt = PSEPT_fmt
cli_support = ['bintocsv', 'csvtobin', 'bintoparquet', 'parquettobin']



#!/usr/bin/env python

import argparse
import logging

from . import manager, logger


def validate_flags(args):
    flags = [
        args.selt,
        args.melt,
        args.qelt,
        args.splt,
        args.mplt,
        args.qplt,
    ]
    if sum(flags) == 0:
        raise RuntimeError("ERROR: no file type flag provided to katpy, run with -h to see which flags are valid")
    if sum(flags) > 1:
        raise RuntimeError("ERROR: katpy cannot kat more than one file type flag")

    if args.files_in is None and args.dir_in is None:
        raise RuntimeError("Error: katpy must specify at least one of --files_in or --dir_in.")

    if args.file_type and args.file_type not in ["csv", "bin"]:
        raise RuntimeError("Error: katpy file_type must be [\"csv\", \"bin\"] or None")


def main():
    parser = argparse.ArgumentParser(description='Concatenate ELT/PLT CSV files')
    parser.add_argument('-o', '--out', type=str, required=True, help='Output Concatenated CSV file')
    parser.add_argument('-f', '--file_type', type=str, default=None, help='Input file type if not discernible from input file suffix.')
    parser.add_argument('-i', '--files_in', type=str, nargs='+', required=False, help='Individual input file paths to concatenate')
    parser.add_argument('-d', '--dir_in', type=str, default=None, help='Path to the directory containing files for concatenation')
    parser.add_argument('-s', '--selt', action='store_true', help='Concatenate SELT CSV file')
    parser.add_argument('-m', '--melt', action='store_true', help='Concatenate MELT CSV file')
    parser.add_argument('-q', '--qelt', action='store_true', help='Concatenate QELT CSV file')
    parser.add_argument('-S', '--splt', action='store_true', help='Concatenate SPLT CSV file')
    parser.add_argument('-M', '--mplt', action='store_true', help='Concatenate MPLT CSV file')
    parser.add_argument('-Q', '--qplt', action='store_true', help='Concatenate QPLT CSV file')
    parser.add_argument('-u', '--unsorted', action='store_true', help='Do not sort by event/period ID')
    parser.add_argument('-v', '--logging-level', type=int, default=30,
                        help='logging level (debug:10, info:20, warning:30, error:40, critical:50)')

    args = parser.parse_args()
    validate_flags(args)
    kwargs = vars(args)

    # Set up logging
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    manager.main(**kwargs)


if __name__ == '__main__':
    main()



import numpy as np

from oasislmf.pytools.common.data import oasis_int, oasis_float


VALID_EXT = ["csv", "bin", "parquet"]

AAL_output = [
    ('SummaryId', oasis_int, '%d'),
    ('SampleType', oasis_int, '%d'),
    ('MeanLoss', oasis_float, '%.6f'),
    ('SDLoss', oasis_float, '%.6f'),
]

AAL_meanonly_output = [
    ('SummaryId', oasis_int, '%d'),
    ('SampleType', oasis_int, '%d'),
    ('MeanLoss', oasis_float, '%.6f'),
]

ALCT_output = [
    ("SummaryId", oasis_float, '%d'),
    ("MeanLoss", oasis_float, '%.6f'),
    ("SDLoss", oasis_float, '%.6f'),
    ("SampleSize", oasis_float, '%d'),
    ("LowerCI", oasis_float, '%.6f'),
    ("UpperCI", oasis_float, '%.6f'),
    ("StandardError", oasis_float, '%.6f'),
    ("RelativeError", oasis_float, '%.6f'),
    ("VarElementHaz", oasis_float, '%.6f'),
    ("StandardErrorHaz", oasis_float, '%.6f'),
    ("RelativeErrorHaz", oasis_float, '%.6f'),
    ("VarElementVuln", oasis_float, '%.6f'),
    ("StandardErrorVuln", oasis_float, '%.6f'),
    ("RelativeErrorVuln", oasis_float, '%.6f'),
]


AAL_headers = [c[0] for c in AAL_output]
AAL_meanonly_headers = [c[0] for c in AAL_meanonly_output]

AAL_dtype = np.dtype([(c[0], c[1]) for c in AAL_output])
AAL_meanonly_dtype = np.dtype([(c[0], c[1]) for c in AAL_meanonly_output])

AAL_fmt = ','.join([c[2] for c in AAL_output])
AAL_meanonly_fmt = ','.join([c[2] for c in AAL_meanonly_output])

ALCT_headers = [c[0] for c in ALCT_output]
ALCT_dtype = np.dtype([(c[0], c[1]) for c in ALCT_output])
ALCT_fmt = ','.join([c[2] for c in ALCT_output])



from oasislmf.pytools.common.data import gul_summary_xref_headers, gul_summary_xref_dtype, gul_summary_xref_fmt


headers = gul_summary_xref_headers
dtype = gul_summary_xref_dtype
fmt = gul_summary_xref_fmt
cli_support = ['bintocsv', 'csvtobin']



from oasislmf.pytools.common.data import returnperiods_headers, returnperiods_dtype, returnperiods_fmt


headers = returnperiods_headers
dtype = returnperiods_dtype
fmt = returnperiods_fmt
cli_support = ['bintocsv', 'csvtobin']
